from ..pysoap.soapenvelope import SoapFault, SoapFaultCode, AdressingFault
class HTTPRequestHandlingError(Exception):
    ''' This class is used to communicate errors from http request handlers back to http server.'''
    def __init__(self, status, reason, soapfault):
        '''
        @param status: integer, e.g. 404
        param reason: the provided human readable text
        '''
        super(HTTPRequestHandlingError, self).__init__()
        self.status = status
        self.reason = reason
        self.soapfault = soapfault

    def __repr__(self):
        if self.soapfault:
            return '{}(status={}, reason={}'.format(self.__class__.__name__, self.status, self.soapfault)
        else:
            return '{}(status={}, reason={}'.format(self.__class__.__name__, self.status, self.reason)


class FunctionNotImplementedError(HTTPRequestHandlingError):
    def __init__(self, request):
        fault = SoapFault(request, code=SoapFaultCode.RECEIVER, reason='not implemented', details='sorry!')
        super().__init__(500, 'not implemented', fault.as_xml())


class InvalidActionError(HTTPRequestHandlingError):
    def __init__(self, request):
        fault = AdressingFault(request,
                               code=SoapFaultCode.SENDER,
                               reason='invalid action {}'.format(request.address.action))
        super().__init__(400, 'Bad Request', fault.as_xml())


class InvalidPathError(HTTPRequestHandlingError):
    def __init__(self, request, path):
        fault = AdressingFault(request,
                               code=SoapFaultCode.SENDER,
                               reason='invalid path {}'.format(path))
        super().__init__(400, 'Bad Request', fault.as_xml())



""" Implementation of some data types used in Participant Model"""
import decimal
from collections import namedtuple
import traceback
import inspect
import itertools
import enum
from lxml import etree as etree_
from sdc11073 import namespaces
from .mdib import containerproperties  as cp
from math import isclose
'''
Interface of  pmTypes:

fromEtreeNode (class method) is a constructor that is used to create a type object from XML node

asEtreeNode: returns an etree node that represents the object
'''


class PropertyBasedPMType(object):
    """ Base class that assumes all data is defined as containerproperties and _props lists all property names."""

    def asEtreeNode(self, qname, nsmap):
        node = etree_.Element(qname, nsmap=nsmap)
        self._updateNode(node)
        return node


    def _updateNode(self, node):
        for prop_name, prop in self._sortedContainerProperties():
            try:
                prop.updateXMLValue(self, node)
            except Exception as ex:
                raise RuntimeError('In {}.{}, {} could not update: {}'.format(self.__class__.__name__, prop_name, str(prop), traceback.format_exc()))


    def updateFromNode(self, node):
        for dummy, prop in self._sortedContainerProperties():
            prop.updateFromNode(self, node)


    def _sortedContainerProperties(self):
        """
        @return: a list of (name, object) tuples of all GenericProperties ( and subclasses)
        list is created based on _props lists of classes
        """
        ret = []
        classes = inspect.getmro(self.__class__)
        for cls in reversed(classes):
            try:
                names = cls._props  # pylint: disable=protected-access
            except:
                continue
            for name in names:
                obj = getattr(cls, name)
                if obj is not None:
                    ret.append((name, obj))
        return ret


    def __eq__(self, other):
        """ compares all properties"""
        try:
            for name, dummy in self._sortedContainerProperties():
                my_value = getattr(self, name)
                other_value = getattr(other, name)
                if my_value == other_value:
                    continue
                elif (isinstance(my_value,  float) or isinstance(other_value,  float)) and isclose(my_value, other_value):
                    continue # float compare (almost equal)
                else:
                    return False
            return True
        except (TypeError, AttributeError):
            return False

    def __ne__(self, other):
        return not self == other

    @classmethod
    def fromNode(cls, node):
        """ default fromNode Constructor that provides no arguments for class __init__"""
        obj = cls()
        obj.updateFromNode(node)
        return obj


class ElementWithTextOnly(PropertyBasedPMType):
    text = cp.NodeTextProperty()  # this is the text of the node. Here attribute is lower case!
    _props = ['text']
    '''An Element that has no attributes, only a text.'''
    def __init__(self, text):
        self.text = text


    @classmethod
    def fromNode(cls, node):
        text = node.text
        return cls(text)    



class LocalizedText(PropertyBasedPMType):
    text = cp.NodeTextProperty()  # this is the text of the node. Here attribute is lower case!
    Ref = cp.NodeAttributeProperty('Ref')
    Lang = cp.NodeAttributeProperty('Lang')
    Version = cp.IntegerAttributeProperty('Version')
    TextWidth = cp.NodeAttributeProperty('TextWidth') # one of xs, s, m, l, xl, xxl
    _props = ['text', 'Ref', 'Lang', 'Version', 'TextWidth']
    ''' Represents a LocalizedText type in the Participant Model. '''
    def __init__(self, text, lang=None, ref=None, version=None, textWidth = None):
        """
        @param text: a string
        @param lang: a string or None
        @param ref: a string or None
        @param version: an int or None
        @param textWidth: xs, s, m, l, xl, xxl or None
        """
        self.text = text
        self.Lang = lang
        self.Ref = ref
        self.Version = version
        self.TextWidth = textWidth


    @classmethod
    def fromNode(cls, node):
        text = node.text
        lang = node.get('Lang')
        ref = node.get('Ref')
        version = node.get('Version')
        if version is not None:
            version = int(version)
        textWidth = node.get('TextWidth')
        return cls(text, lang, ref, version, textWidth)


        
DefaultCodingSystem = 'urn:oid:1.2.840.10004.1.1.1.0.0.1' # ISO/IEC 11073-10101


def _get_default_coding_system():
    """Helper method that allows handling of DefaultCodingSystem modification at runtime.
    By using it implied coding system always is the current value of DefaultCodingSystem."""
    return DefaultCodingSystem


class NotCompareableVersionError(Exception):
    """ This exception says that two coded values cannot be compared, because one has a coding system version, the other one not.
    In that case it is not possible to decide if they are equal."""
    pass

_CodingBase = namedtuple('_CodingBase', 'code codingSystem codingSystemVersion')
class Coding(_CodingBase):
    """ Immutable representation of a coding. Can be used as key in dictionaries"""


    def __new__(cls, code, codingSystem=None, codingSystemVersion=None):
        return super(Coding, cls).__new__(cls,
                                          str(code),
                                          codingSystem or DefaultCodingSystem,
                                          codingSystemVersion)

    def equals(self, other, raiseNotComparableException=False):
        """ different compare method to __eq__, overwriting this one makes Coding unhashable!
         other can be an int, a string, or a Coding.
         Simple comparison with int or string only works if self.codingSystem == DefaultCodingSystem
         and self.codingSystemVersion is None"""
        if isinstance(other, int):
            other = str(other)
        if isinstance(other, str):
            # compare to 11073 coding system where codes are strings
            if self.codingSystem != DefaultCodingSystem:
                return False
            elif self.code != other:
                return False
            elif self.codingSystemVersion is None:
                return True
            else:
                if raiseNotComparableException:
                    raise NotCompareableVersionError('no simple compare, self.codingSystemVersion == {}'.format(self.codingSystemVersion))
                else:
                    return False
        else:
            try:
                if self.code != other.code or self.codingSystem != other.codingSystem:
                    return False
                elif self.codingSystemVersion == other.codingSystemVersion:
                    return True
                elif self.codingSystemVersion is None or other.codingSystemVersion is None:
                    if raiseNotComparableException:
                        raise NotCompareableVersionError('my codingSystem = "{}", other = "{}"'.format(self.codingSystemVersion, other.codingSystemVersion))
                    else:
                        return False
                else:
                    return False
            except AttributeError:
                return False

    @classmethod
    def fromNode(cls, node):
        """ Read Code and CodingSystem attributes of a node (CodedValue). """
        code = node.get('Code')
        codingSystem = node.get('CodingSystem', DefaultCodingSystem)
        codingSystemVersion = node.get('CodingSystemVersion')
        return cls(code, codingSystem, codingSystemVersion)


def mkCoding(code, codingSystem=None, codingSystemVersion=None):
    return Coding(code, codingSystem or DefaultCodingSystem, codingSystemVersion)



class T_Translation(PropertyBasedPMType):
    """
    Translation is part of CodedValue in BICEPS FINAL
    """
    ext_Extension = cp.ExtensionNodeProperty()
    Code = cp.NodeAttributeProperty('Code')
    CodingSystem = cp.NodeAttributeProperty('CodingSystem', impliedPyValue=_get_default_coding_system)
    CodingSystemVersion = cp.NodeAttributeProperty('CodingSystemVersion')

    _props = ['ext_Extension', 'Code', 'CodingSystem', 'CodingSystemVersion']

    def __init__(self, code=None, codingsystem=None, codingSystemVersion=None):
        """

        @param code: a string or an int
        @param codingsystem: anyURI or None, defaults to ISO/IEC 11073-10101 if None
        @param codingSystemVersion: a string, min. length = 1
        """
        self.Code = str(code)
        self.CodingSystem = codingsystem
        self.CodingSystemVersion = codingSystemVersion
        self.coding = None # oberwritten by self._mkCoding()
        self.mkCoding()

    def __repr__(self):
        if self.CodingSystem is None:
            return 'CodedValue({})'.format(self.Code)
        elif self.CodingSystemVersion is None:
            return 'CodedValue({}, codingsystem={})'.format(self.Code, self.CodingSystem)
        else:
            return 'CodedValue({}, codingsystem={}, codingsystemversion={})'.format(self.Code, self.CodingSystem, self.CodingSystemVersion)

    def mkCoding(self):
        if self.Code is not None:
            self.coding = Coding(self.Code, self.CodingSystem, self.CodingSystemVersion)
        else:
            self.coding = None

    def __eq__(self, other):
        """ other can be an int, a string, a CodedValue like object (has "coding" member) or a Coding"""
        if hasattr(other, 'coding'):
            return self.coding.equals(other.coding)
        else:
            return self.coding.equals(other)

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        obj.updateFromNode(node)
        obj.mkCoding()
        return obj


class _CodedValueBase(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    CodingSystemName = cp.SubElementListProperty([namespaces.domTag('CodingSystemName')], cls = LocalizedText)
    ConceptDescription = cp.SubElementListProperty([namespaces.domTag('ConceptDescription')], cls = LocalizedText)
    Code = cp.NodeAttributeProperty('Code')
    CodingSystem = cp.NodeAttributeProperty('CodingSystem', impliedPyValue=_get_default_coding_system)
    CodingSystemVersion = cp.NodeAttributeProperty('CodingSystemVersion')
    SymbolicCodeName = cp.NodeAttributeProperty('SymbolicCodeName')
    Type = cp.XsiTypeAttributeProperty(namespaces.QN_TYPE)
    _props = ['ext_Extension', 'CodingSystemName', 'ConceptDescription',
              'Code', 'CodingSystem', 'CodingSystemVersion', 'SymbolicCodeName', 'Type']
    # aliases for backward compatibility
    codingSystemNames = CodingSystemName
    conceptDescriptions = ConceptDescription
    codingSystem = CodingSystem

    def __init__(self, code, codingsystem=None, codingSystemVersion=None, codingSystemNames=None, conceptDescriptions=None, symbolicCodeName=None):
        """
        @param code: a string or an int
        @param codingsystem: anyURI or None, defaults to ISO/IEC 11073-10101 if None
        @param codingSystemVersion: a string, min. length = 1
        @param codingSystemNames: a list of LocalizedText objects or None
        @param conceptDescriptions: a list of LocalizedText objects or None
        @param symbolicCodeName: a string, min. length = 1 or None
        """
        self.Code = str(code)
        self.CodingSystem = codingsystem
        self.CodingSystemVersion = codingSystemVersion
        self.CodingSystemName = [] if codingSystemNames is None else codingSystemNames
        self.ConceptDescription = [] if conceptDescriptions is None else conceptDescriptions
        self.SymbolicCodeName = symbolicCodeName
        self.coding = None # oberwritten by self._mkCoding()
        self.mkCoding()


    def mkCoding(self):
        if self.Code is not None:
            self.coding = Coding(self.Code, self.CodingSystem, self.CodingSystemVersion)
        else:
            self.coding = None

    def __repr__(self):
        if self.CodingSystem is None:
            return 'CodedValue({})'.format(self.Code)
        elif self.CodingSystemVersion is None:
            return 'CodedValue({}, codingsystem="{}")'.format(self.Code, self.CodingSystem)
        else:
            return 'CodedValue({}, codingsystem="{}", codingsystemversion="{}")'.format(self.Code, self.CodingSystem, self.CodingSystemVersion)

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        obj.updateFromNode(node)
        obj.mkCoding()
        return obj



class CodedWithTranslations(_CodedValueBase):
    Translation = cp.SubElementListProperty([namespaces.domTag('Translation')], cls = _CodedValueBase)
    _props = ['Translation']

    def __eq__(self, other):
        """ other can be an int, a string, a CodedValue like object (has "coding" member) or a Coding"""
        if hasattr(other, 'coding'):
            return self.coding == other.coding
        else:
            return self.coding == other



class CodedValue(_CodedValueBase):
    Translation = cp.SubElementListProperty([namespaces.domTag('Translation')], cls = T_Translation)
    _props = ['Translation']

    def __eq__(self, other):
        """ This operator handles not comparable versions as different versions

        :param other: int, str or another CodedValue
        """
        return self.equals(other, raiseNotComparableException=False)


    def equals(self, other, raiseNotComparableException=True):
        """
        Compare this CodedValue with another one.
        A simplified compare with an int or string is also possible, it assumes DefaultCodingSystem and no CodingSystemVersion
        :param other: int, str or another CodedValue
        :param raiseNotComparableException: if False, not comparable versions are handled as different versions
                        if True, a NotCompareableVersionError is thrown if any of the Codings are not comparable
        :return: boolean
        """
        if isinstance(other, self.__class__):
                # Two CodedValue objects C1 and C2 are equivalent, if there exists a CodedValue object T1 in C1/pm:Translation
                # and a CodedValue object T2 in C2/pm:Translation such that T1 and T2 are equivalent, C1 and T2 are equivalent, or C2 and T1 are equivalent.
                foundMatch = False
                not_comparables = []
                my_codes = [self] # C1
                my_codes.extend(self.Translation) # all T1
                other_codes = [other]   #C2
                other_codes.extend(other.Translation) # all T2
                for left, right in itertools.product(my_codes, other_codes):
                    try:
                        if left.coding.equals(right.coding, raiseNotComparableException):
                            if not raiseNotComparableException:
                                return True
                            else:
                                foundMatch = True
                    except NotCompareableVersionError as ex:
                        not_comparables.append(str(ex))
                if not_comparables:
                    raise NotCompareableVersionError(';'.join(not_comparables))
                else:
                    return foundMatch
        else:
            # simplified compare: compare to 11073 coding system where codes are integers, ignore translations
            return self.coding.equals(other, raiseNotComparableException)

    @classmethod
    def fromNode(cls, node):
        nodeType = None
        nodeTypeText = node.get(namespaces.QN_TYPE)
        if nodeTypeText is not None:
            nodeType = namespaces.txt2QName(node.get(namespaces.QN_TYPE), node.nsmap)
        if nodeType == namespaces.domTag('CodedWithTranslations'):
            cls = CodedWithTranslations  # pylint: disable=self-cls-assignment

        obj = cls(None)
        obj.updateFromNode(node)
        obj.mkCoding()
        return obj


class Annotation(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    _props = ['ext_Extension', 'Type']
    
    codedValue = Type
    ''' An Annotation contains a Type Element that is a CodedValue.
    This is intended as an immutable object. After it has been created, no modification shall be done. '''
    def __init__(self, codedValue):
        self.Type = codedValue
        self.coding = codedValue.coding
 
 
    @classmethod
    def fromNode(cls, node):
        typeNode = node.find(namespaces.domTag('Type'))
        codedValue = CodedValue.fromNode(typeNode)
        return cls(codedValue)


class OperationGroup(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    OperatingMode = cp.NodeAttributeProperty('OperatingMode')
    Operations = cp.NodeAttributeListProperty('Operations') #pm:OperationRef
    _props = ['ext_Extension', 'Type', 'OperatingMode', 'Operations']

    def __init__(self, codedValue=None, operatingMode=None, operations=None):
        """
        @param codedValue: a CodedValue instances or None
        @param operatingMode:  xsd:string string
        @param operations: a xsd:string
        """
        self.Type = codedValue
        self.OperatingMode = operatingMode
        self.Operations = operations

    @classmethod
    def fromNode(cls, node):
        ret = cls()
        ret.updateFromNode(node)
        return ret


class InstanceIdentifier(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    IdentifierName = cp.SubElementListProperty([namespaces.domTag('IdentifierName')], cls = LocalizedText)
    Root = cp.NodeAttributeProperty('Root', defaultPyValue='biceps.uri.unk') # xsd:anyURI string, default is defined in R0135
    Extension = cp.NodeAttributeProperty('Extension') # a xsd:string
    _props=('ext_Extension', 'Type', 'IdentifierName', 'Root', 'Extension')

    def __init__(self, root, type_codedValue=None, identifierNames=None, extensionString=None):
        """
        @param root:  xsd:anyURI string
        @param type_codedValue: a CodedValue instances or None
        @param identifierNames: a list of LocalizedText instances or None
        @param extensionString: a xsd:string
        """
        self.Root = root
        self.Type = type_codedValue
        self.IdentifierName = [] if identifierNames is None else identifierNames
        self.Extension = extensionString
        self.node = None


    @classmethod
    def fromNode(cls, node):
        root = None
        extensionString = None
        type_codedValue = None
        identifierNames = None
        ret = cls(root, type_codedValue=type_codedValue, identifierNames=identifierNames, extensionString=extensionString)
        ret.updateFromNode(node)
        ret.node = node 
        return ret


    def __repr__(self):
        return 'InstanceIdentifier(root={!r}, Type={} ext={!r})'.format(self.Root, self.Type, self.Extension)



class Range(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Lower = cp.DecimalAttributeProperty('Lower') # optional, an integer or float
    Upper = cp.DecimalAttributeProperty('Upper') # optional, an integer or float
    StepWidth = cp.DecimalAttributeProperty('StepWidth') # optional, an integer or float
    RelativeAccuracy = cp.DecimalAttributeProperty('RelativeAccuracy') # optional, an integer or float
    AbsoluteAccuracy = cp.DecimalAttributeProperty('AbsoluteAccuracy') # optional, an integer or float
    _props= ['ext_Extension', 'Lower', 'Upper', 'StepWidth', 'RelativeAccuracy', 'AbsoluteAccuracy']
    
    def __init__(self, lower=None, upper=None, stepWidth=None, relativeAccuracy=None, absoluteAccuracy=None):
        """
        @param lower: The including lower bound of the range. A value as float or integer, can be None
        @param upper: The including upper bound of the range. A value as float or integer, can be None
        @param stepWidth: The numerical distance between two values in the range of the given upper and lower bound. A value as float or integer, can be None
        @param relativeAccuracy: Maximum relative error in relation to the correct value within the given range. A value as float or integer, can be None
        @param absoluteAccuracy: Maximum absolute error in relation to the correct value within the given range. A value as float or integer, can be None
        """
        self.Lower = lower
        self.Upper = upper
        self.StepWidth = stepWidth
        self.RelativeAccuracy = relativeAccuracy
        self.AbsoluteAccuracy = absoluteAccuracy
        
        
    def __repr__(self):
        return 'Range (Lower={!r}, Upper={!r}, StepWidth={!r}, RelativeAccuracy={!r}, AbsoluteAccuracy={!r})'\
            .format(self.Lower, self.Upper, self.StepWidth, self.RelativeAccuracy, self.AbsoluteAccuracy)



class Measurement(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    MeasurementUnit = cp.SubElementProperty([namespaces.domTag('MeasurementUnit')], valueClass=CodedValue) # mandatory
    MeasuredValue = cp.DecimalAttributeProperty('MeasuredValue')
    _props = ['ext_Extension', 'MeasurementUnit', 'MeasuredValue']
    
    def __init__(self, value, unit):
        """
        @param value: a value as string, float or integer
        @param unit: a CodedValue instance
        """
        self.MeasuredValue = value
        self.MeasurementUnit = unit
    
    
    @classmethod
    def fromNode(cls, node):
        value = node.get('MeasuredValue')
        if value is not None:
            value = float(value)
        unit = None
        unitNode = node.find(namespaces.domTag('MeasurementUnit'))
        if unitNode is not None:
            unit = CodedValue.fromNode(unitNode)
        return cls(value, unit)


    def __repr__(self):
        return 'Measurement(value={!r}, Unit={!r})'.format(self.MeasuredValue, self.MeasurementUnit)



class AllowedValue(PropertyBasedPMType):
    Value = cp.NodeTextProperty([namespaces.domTag('Value')])
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue)
    _props=['Value', 'Type']
    typeCoding = Type
    value = Value
    
    def __init__(self, valueString, typeCoding=None):
        """One AllowedValue of a EnumStringMetricDescriptor. It has up to two sub elements "Value" and "Type"(optional).
        A StringEnumMetricDescriptor has a list of AllowedValues.
        @param valueString: a string
        @param typeCoding: an optional CodedValue instance
        """
        self.Value = valueString
        self.Type = typeCoding


    @classmethod
    def fromNode(cls, node):
        valueString = node.find(namespaces.domTag('Value')).text
        typeNode = node.find(namespaces.domTag('Type'))
        if typeNode is None:
            typeCoding = None
        else:
            typeCoding = CodedValue.fromNode(typeNode)
        return cls(valueString, typeCoding)

    

class AbstractMetricValue(PropertyBasedPMType):
    """ This is the base class for metric values inside metric states"""
    ext_Extension = cp.ExtensionNodeProperty()
    StartTime = cp.TimestampAttributeProperty('StartTime') # time.time() value (float)
    StopTime = cp.TimestampAttributeProperty('StopTime') # time.time() value (float)
    DeterminationTime = cp.TimestampAttributeProperty('DeterminationTime') # time.time() value (float)
    MQ_Extension = cp.ExtensionNodeProperty([namespaces.domTag('MetricQuality')])
    Validity = cp.NodeAttributeProperty('Validity', [namespaces.domTag('MetricQuality')]) # pm:MeasurementValidity
    Mode = cp.NodeAttributeProperty('Mode', [namespaces.domTag('MetricQuality')], impliedPyValue='Real') # pm:GenerationMode
    Qi = cp.DecimalAttributeProperty('Qi', [namespaces.domTag('MetricQuality')], impliedPyValue=1) # pm:QualityIndicator
    Annotation = cp.SubElementListProperty([namespaces.domTag('Annotation')], Annotation)
    _props = ('ext_Extension', 'StartTime', 'StopTime', 'DeterminationTime', 'MQ_Extension', 'Validity', 'Mode', 'Qi', 'Annotation')
    
    Annotations = Annotation
    def __init__(self, nsmapper, node=None):
        self._nsmapper = nsmapper
        # attributes of root node
        self.node = node
        if node is not None:
            self.updateFromNode(node)
        else:
            self.Validity = 'Vld' # mandatory value, for convenience it is preset to Vld


    def updateFromNode(self, node):
        for dummy, prop in self._sortedContainerProperties():
            prop.updateFromNode(self, node)
        self.node = node    


    def asEtreeNode(self, qname, nsmap):
        node = super(AbstractMetricValue, self).asEtreeNode(qname, nsmap)
        node.set(namespaces.QN_TYPE, namespaces.docNameFromQName(self.QType, nsmap))
        return node


    @classmethod
    def fromNode(cls, node):
        obj = cls(node.nsmap, node)
        return obj



class NumericMetricValue(AbstractMetricValue):
    QType = namespaces.domTag('NumericMetricValue')
    Value = cp.DecimalAttributeProperty('Value') # an integer or float
    _props = ('Value',)    


    def __repr__(self):
        return '{} Validity={} Value={} DeterminationTime={}'.format(self.__class__.__name__,
                                                                     self.Validity,
                                                                     self.Value,
                                                                     self.DeterminationTime)

    def __eq__(self, other):
        """ compares all properties, special handling of Value member"""
        try:
            for name, dummy in self._sortedContainerProperties():
                if name == 'Value':
                    my_value = getattr(self, name)
                    other_value = getattr(other, name)
                    if my_value is None and other_value is None:
                        continue
                    elif my_value is None or other_value is None:
                        return False  # only one is None, they are not equal
                    else:
                        if decimal.Decimal(str(my_value)) != decimal.Decimal(str(other_value)):
                            return False
                else:
                    if getattr(self, name) == getattr(other, name):
                        continue
                    else:
                        return False
            return True
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other



class StringMetricValue(AbstractMetricValue):
    QType = namespaces.domTag('StringMetricValue')
    Value = cp.NodeAttributeProperty('Value')  # a string
    _props = ('Value',)    
    

    def __repr__(self):
        return '{} Validity={} Value={} DeterminationTime={}'.format(self.__class__.__name__,
                                                                     self.Validity,
                                                                     self.Value,
                                                                     self.DeterminationTime)



class ApplyAnnotation(PropertyBasedPMType):
    AnnotationIndex = cp.IntegerAttributeProperty('AnnotationIndex')
    SampleIndex = cp.IntegerAttributeProperty('SampleIndex')
    _props = ['AnnotationIndex', 'SampleIndex']

    def __init__(self, annotationIndex, sampleIndex):
        self.AnnotationIndex = annotationIndex
        self.SampleIndex = sampleIndex   

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.AnnotationIndex.updateFromNode(obj, node)
        cls.SampleIndex.updateFromNode(obj, node)
        return obj

    def __repr__(self):
        return '{} AnnotationIndex={} SampleIndex={}'.format(self.__class__.__name__, self.AnnotationIndex, self.SampleIndex)



class SampleArrayValue(AbstractMetricValue):
    QType = namespaces.domTag('SampleArrayValue')
    Samples = cp.DecimalListAttributeProperty('Samples') # list of xs:decimal types 
    ApplyAnnotations = cp.SubElementListProperty([namespaces.domTag('ApplyAnnotation')], ApplyAnnotation)
    _props = ('Samples', 'ApplyAnnotations')    


    def __repr__(self):
        return '{} Samples={} ApplyAnnotations={}'.format(self.__class__.__name__, self.Samples, self.ApplyAnnotations)


    def __eq__(self, other):
        """ compares all properties, special handling of Value member"""
        try:
            for name, dummy in self._sortedContainerProperties():
                if name == 'Samples':
                    own_sample = getattr(self, name)
                    other_sample = getattr(other, name)
                    # avoid None values; treat None and empty list as equal
                    if own_sample is None:
                        own_sample = []
                    if other_sample is None:
                        other_sample = []
                    if len(own_sample) != len(other_sample):
                        return False
                    for pair in zip(own_sample, other_sample):
                        if decimal.Decimal(str(pair[0])) != decimal.Decimal(str(pair[1])):
                            return False
                else:
                    if getattr(self, name) == getattr(other, name):
                        continue
                    else:
                        return False
            return True
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other


class RemedyInfo(PropertyBasedPMType):
    """An Element that has
         0..1 Subelement "Extension" (not handled here)
         0..n SubElements "Description" type=pm:LocalizedText."""
    ext_Extension = cp.ExtensionNodeProperty()
    Description = cp.SubElementListProperty([namespaces.domTag('Description')], cls = LocalizedText)
    _props = ['ext_Extension', 'Description']

    def __init__(self, descriptions=None):
        """
        @param descriptions : a list of LocalizedText objects or None
        """
        if descriptions:
            self.Description = descriptions
        


class CauseInfo(PropertyBasedPMType):
    """An Element that has
         0..1 Subelement "RemedyInfo", type = pm:RemedyInfo
         0..n SubElements "Description" type=pm:LocalizedText."""
    ext_Extension = cp.ExtensionNodeProperty()
    RemedyInfo = cp.SubElementProperty([namespaces.domTag('RemedyInfo')], valueClass=RemedyInfo)
    Description = cp.SubElementListProperty([namespaces.domTag('Description')], cls = LocalizedText)
    _props = ['ext_Extension', 'RemedyInfo', 'Description']
    def __init__(self, remedyInfo, descriptions):
        """
        @param remedyInfo: a RemedyInfo instance or None
        @param descriptions : a list of LocalizedText objects or None
        """
        self.RemedyInfo = remedyInfo
        self.Description = descriptions


    @classmethod
    def fromNode(cls, node):
        remedyInfoNode = node.find(namespaces.domTag('RemedyInfo'))
        if remedyInfoNode is not None:
            remedyInfo = RemedyInfo.fromNode(remedyInfoNode)
        else:
            remedyInfo = None
        descriptions = []
        descriptionNodes = node.findall(namespaces.domTag('Description'))
        for d in descriptionNodes:
            descriptions.append(LocalizedText.fromNode(d))
        return cls(remedyInfo, descriptions)    



class Argument(PropertyBasedPMType):
    """An Element that has
         1 Subelement "ArgName", type = pm:CodedValue
         1 SubElement "Arg" type=QName."""
    ArgName = cp.SubElementProperty([namespaces.domTag('ArgName')], valueClass=CodedValue)
    Arg = cp.NodeTextQNameProperty([namespaces.domTag('Arg')])
    _props = ['ArgName', 'Arg']
    def __init__(self, argName, arg):
        """
        @param argName: a CodedValue instance
        @param arg : etree_.QName instance
        """
        self.ArgName = argName
        self.Arg = arg
        

    @classmethod
    def fromNode(cls, node):
        argNameNode = node.find(namespaces.domTag('ArgName'))
        argName = CodedValue.fromNode(argNameNode)
        argNode = node.find(namespaces.domTag('Arg'))
        arg_QName = namespaces.txt2QName(argNode.text, node.nsmap)
        return cls(argName, arg_QName)    


    def __repr__(self):
        return 'Argument(argName={}, arg={})'.format(self.ArgName, self.Arg)



class PhysicalConnectorInfo(PropertyBasedPMType):
    """PhysicalConnectorInfo defines a number in order to allow to guide the clinical user for a failure,
    e.g., in case of a disconnection of a sensor or an ultrasonic handpiece.
    Only in BICEPS final!"""
    ext_Extension = cp.ExtensionNodeProperty()
    Label = cp.SubElementListProperty([namespaces.domTag('Label')], cls = LocalizedText) # A human-readable label that describes the physical connector.
    Number = cp.IntegerAttributeProperty('Number')# Number designates the connector number of the physical connector.
    _props = ['ext_Extension', 'Label', 'Number']

    def __init__(self, labels, number):
        """
        @param labels: a  list of LocalizedText
        @param number : an integer
        """
        self.Label = labels
        self.Number = number

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.Label.updateFromNode(obj, node)
        cls.Number.updateFromNode(obj, node)
        return obj

    def __repr__(self):
        return 'PhysicalConnectorInfo(label={}, number={})'.format(self.Label, self.Number)


class SystemSignalActivation(PropertyBasedPMType):
    Manifestation = cp.NodeAttributeProperty('Manifestation', defaultPyValue='Oth') # required, pmtypes.AlertSignalManifestation
    State = cp.NodeAttributeProperty('State', defaultPyValue='On')                 # required,  pmtypes.AlertActivation
    _props = ['Manifestation', 'State']

    def __init__(self, manifestation, state):
        """
        @param manifestation: a pmtypes.AlertSignalManifestation value
        @param state : a pmtypes.AlertActivation value
        """
        self.Manifestation = manifestation
        self.State = state

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        obj.updateFromNode(node)
        return obj

    def __repr__(self):
        return 'SystemSignalActivation(Manifestation={}, State={})'.format(self.Manifestation, self.State)


class ProductionSpecification(PropertyBasedPMType):
    SpecType = cp.SubElementProperty([namespaces.domTag('SpecType')], valueClass=CodedValue)
    ProductionSpec = cp.NodeTextProperty([namespaces.domTag('ProductionSpec')], isOptional=False)
    ComponentId = cp.SubElementProperty([namespaces.domTag('ComponentId')], valueClass=InstanceIdentifier) # optional
    _props = ['SpecType', 'ProductionSpec', 'ComponentId']

    def __init__(self, spectype, productionspec, componentid=None):
        """
        @param spectype: a pmtypes.CodedValue value
        @param productionspec: a string
        @param componentid : a pmtypes.InstanceIdentifier value
        """
        self.SpecType = spectype
        self.ProductionSpec = productionspec
        self.ComponentId = componentid

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        obj.updateFromNode(node)
        return obj


class BaseDemographics(PropertyBasedPMType):
    Givenname = cp.NodeTextProperty([namespaces.domTag('Givenname')])
    Middlename = cp.SubElementListProperty([namespaces.domTag('Middlename')], cls=ElementWithTextOnly)  # 0...n Elements
    Familyname = cp.NodeTextProperty([namespaces.domTag('Familyname')])
    Birthname = cp.NodeTextProperty([namespaces.domTag('Birthname')])
    Title = cp.NodeTextProperty([namespaces.domTag('Title')])
    _props = ['Givenname', 'Middlename', 'Familyname', 'Birthname', 'Title']

    def __init__(self, givenname=None, middlenames=None, familyname=None, birthname=None, title=None):
        self.Givenname = givenname
        if isinstance(middlenames, str):
            self.Middlename = [middlenames]
        else:
            self.Middlename = middlenames or []
        self.Familyname = familyname
        self.Birthname = birthname
        self.Title = title


class PersonReference(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    Identification = cp.SubElementListProperty([namespaces.domTag('Identification')], cls=InstanceIdentifier) # 1...n
    Name = cp.SubElementProperty([namespaces.domTag('Name')], valueClass=BaseDemographics) # optional
    _props = ['ext_Extension', 'Identification', 'Name']

    def __init__(self, identifications=None, name=None):
        """
        :param identifications: a list of InstanceIdentifier objects
        :param name: a BaseDemographics object
        """
        if identifications:
            self.Identification = identifications
        self.Name = name


class LocationDetail(PropertyBasedPMType):
    ext_Extension = cp.ExtensionNodeProperty()
    PoC = cp.NodeAttributeProperty('PoC')
    Room = cp.NodeAttributeProperty('Room')
    Bed = cp.NodeAttributeProperty('Bed')
    Facility = cp.NodeAttributeProperty('Facility')
    Building = cp.NodeAttributeProperty('Building')
    Floor = cp.NodeAttributeProperty('Floor')
    _props = ('ext_Extension', 'PoC', 'Room', 'Bed', 'Facility', 'Building', 'Floor')

    def __init__(self, poc=None, room=None, bed=None, facility=None, building=None, floor=None):
        self.PoC = poc
        self.Room = room
        self.Bed = bed
        self.Facility = facility
        self.Building = building
        self.Floor = floor


class LocationReference(PropertyBasedPMType): # BICEPS Final
    Identification = cp.SubElementListProperty([namespaces.domTag('Identification')], cls=InstanceIdentifier) # 1...n
    LocationDetail = cp.SubElementProperty([namespaces.domTag('LocationDetail')], valueClass=LocationDetail) # optional
    _props = ['Identification', 'LocationDetail']

    def __init__(self, identifications=None, locationdetail=None):
        if identifications:
            self.Identification = identifications
        self.LocationDetail = locationdetail


class PersonParticipation(PersonReference):
    Role = cp.SubElementListProperty([namespaces.domTag('Role')], cls=CodedValue) #0...n
    _props = ['Role',]

    def __init__(self, identifications=None, name=None, roles=None):
        super(PersonParticipation, self).__init__(identifications, name)
        if roles:
            self.Role = roles


class ClinicalInfo(PropertyBasedPMType):
    Type = cp.SubElementProperty([namespaces.domTag('Type')], valueClass=CodedValue) # optional
    Description = cp.SubElementListProperty([namespaces.domTag('Description')], cls=LocalizedText) #0...n
    RelatedMeasurement = cp.SubElementListProperty([namespaces.domTag('RelatedMeasurement')], cls=Measurement) #0...n
    _props = ['Type', 'Description', 'RelatedMeasurement']

    def __init__(self, typecode=None, descriptions=None, relatedmeasurements=None):
        """
        :param typecode: a CodedValue Instance
        :param descriptions: a list of LocalizedText objects
        :param relatedmeasurements: a list of Measurement objects
        """
        self.Type = typecode
        if descriptions:
            self.Description = descriptions
        if relatedmeasurements:
            self.RelatedMeasurement = relatedmeasurements


class ImagingProcedure(PropertyBasedPMType):
    AccessionIdentifier = cp.SubElementProperty([namespaces.domTag('AccessionIdentifier')], valueClass=InstanceIdentifier) # mandatory
    RequestedProcedureId = cp.SubElementProperty([namespaces.domTag('RequestedProcedureId')], valueClass=InstanceIdentifier) # mandatory
    StudyInstanceUid = cp.SubElementProperty([namespaces.domTag('StudyInstanceUid')], valueClass=InstanceIdentifier) # mandatory
    ScheduledProcedureStepId = cp.SubElementProperty([namespaces.domTag('ScheduledProcedureStepId')], valueClass=InstanceIdentifier) # mandatory
    Modality = cp.SubElementProperty([namespaces.domTag('Modality')], valueClass=CodedValue) # optional
    ProtocolCode = cp.SubElementProperty([namespaces.domTag('ProtocolCode')], valueClass=CodedValue) # optional
    _props = ['AccessionIdentifier', 'RequestedProcedureId', 'StudyInstanceUid', 'ScheduledProcedureStepId',
              'Modality', 'ProtocolCode']

    def __init__(self, accessionidentifier, requestedprocedureid, studyinstanceuid, scheduledprocedurestepid,
                 modality=None, protocolcode=None):
        self.AccessionIdentifier = accessionidentifier
        self.RequestedProcedureId = requestedprocedureid
        self.StudyInstanceUid = studyinstanceuid
        self.ScheduledProcedureStepId = scheduledprocedurestepid
        self.Modality = modality
        self.ProtocolCode = protocolcode

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None, None, None)
        obj.updateFromNode(node)
        return obj


class OrderDetail(PropertyBasedPMType):
    Start = cp.NodeTextProperty([namespaces.domTag('Start')]) # optional, xsd:dateTime
    End = cp.NodeTextProperty([namespaces.domTag('End')]) # optional, xsd:dateTime
    Performer = cp.SubElementListProperty([namespaces.domTag('Performer')], cls=PersonParticipation) #0...n)
    Service = cp.SubElementListProperty([namespaces.domTag('Service')], cls=CodedValue) #0...n)
    ImagingProcedure = cp.SubElementListProperty([namespaces.domTag('ImagingProcedure')], cls=ImagingProcedure)
    _props = ['Start', 'End', 'Performer', 'Service', 'ImagingProcedure']

    def __init__(self, start=None, end=None, performer=None, service=None, imagingprocedure=None):
        """
        :param start: a xsd:DateTime string
        :param end: a xsd:DateTime string
        :param performer: a list of PersonParticipation objects
        :param service: a list of CodedValue objects
        @param imagingprocedure: a list of ImagingProcedure objects
        """
        self.Start = start
        self.End = end
        if performer:
            self.Performer = performer
        if service:
            self.Service = service
        if imagingprocedure:
            self.ImagingProcedure = imagingprocedure


class RequestedOrderDetail(OrderDetail):
    ReferringPhysician = cp.SubElementProperty([namespaces.domTag('ReferringPhysician')], valueClass=PersonReference) # optional
    RequestingPhysician = cp.SubElementProperty([namespaces.domTag('RequestingPhysician')], valueClass=PersonReference) # optional
    PlacerOrderNumber = cp.SubElementProperty([namespaces.domTag('PlacerOrderNumber')], valueClass=InstanceIdentifier) # mandatory
    _props = ['ReferringPhysician', 'RequestingPhysician', 'PlacerOrderNumber']


    def __init__(self, start=None, end=None, performer=None, service=None, imagingprocedure=None,
                 referringphysician=None, requestingphysician=None, placerordernumber=None):
        """
        :param referringphysician:  a PersonReference
        :param requestingphysician: a PersonReference
        :param placerordernumber:   an InstanceIdentifier
        """
        super(RequestedOrderDetail, self).__init__(start, end, performer, service, imagingprocedure)
        self.ReferringPhysician = referringphysician
        self.RequestingPhysician = requestingphysician
        self.PlacerOrderNumber = placerordernumber


class PerformedOrderDetail(OrderDetail):
    FillerOrderNumber = cp.SubElementProperty([namespaces.domTag('FillerOrderNumber')], valueClass=InstanceIdentifier) # optional
    ResultingClinicalInfo = cp.SubElementListProperty([namespaces.domTag('RelevantClinicalInfo')], cls=ClinicalInfo)
    _props = ['FillerOrderNumber', 'ResultingClinicalInfo']

    def __init__(self, start=None, end=None, performer=None, service=None, imagingprocedure=None,
                 fillerordernumber=None, resultingclinicalinfos=None):
        super(PerformedOrderDetail, self).__init__(start, end, performer, service, imagingprocedure)
        self.FillerOrderNumber = fillerordernumber
        if resultingclinicalinfos:
            self.ResultingClinicalInfo = resultingclinicalinfos



class WorkflowDetail(PropertyBasedPMType):
    Patient = cp.SubElementProperty([namespaces.domTag('Patient')], valueClass=PersonReference) # optional
    AssignedLocation = cp.SubElementProperty([namespaces.domTag('AssignedLocation')], valueClass=LocationReference) # optional
    VisitNumber = cp.SubElementProperty([namespaces.domTag('VisitNumber')], valueClass=InstanceIdentifier) # optional
    DangerCode = cp.SubElementListProperty([namespaces.domTag('Reason')], cls=CodedValue)
    RelevantClinicalInfo = cp.SubElementListProperty([namespaces.domTag('RelevantClinicalInfo')], cls=ClinicalInfo)
    RequestedOrderDetail = cp.SubElementProperty([namespaces.domTag('RequestedOrderDetail')], valueClass=RequestedOrderDetail) # optional
    PerformedOrderDetail = cp.SubElementProperty([namespaces.domTag('PerformedOrderDetail')], valueClass=PerformedOrderDetail) # optional
    _props = ['Patient', 'AssignedLocation', 'VisitNumber', 'DangerCode',
              'RelevantClinicalInfo', 'RequestedOrderDetail', 'PerformedOrderDetail']


    def __init__(self, patient=None, assignedlocation=None, visitnumber=None, dangercode=None,
                 relevantclinicalinfo=None, requestedorderdetail=None, performedorderdetail=None):
        self.Patient = patient
        self.AssignedLocation = assignedlocation
        self.VisitNumber = visitnumber
        if dangercode:
            self.DangerCode = dangercode
        if relevantclinicalinfo:
            self.RelevantClinicalInfo = relevantclinicalinfo
        self.RequestedOrderDetail = requestedorderdetail
        self.PerformedOrderDetail = performedorderdetail



class Relation(PropertyBasedPMType):
    """Relation allows the modelling of relationships between a metric and other containtment tree entries."""
    Code = cp.SubElementProperty([namespaces.domTag('Code')], valueClass=CodedValue) #optional
    Identification = cp.SubElementProperty([namespaces.domTag('Identification')], valueClass=InstanceIdentifier) # optional
    Kind = cp.NodeAttributeProperty('Kind') # required, Rcm, PS, SST, ECE, DCE, Oth
    Entries = cp.NodeAttributeListProperty('Entries')
    _props = ['Code', 'Identification', 'Kind', 'Entries']


# SafetyReq definitions
class T_Selector(PropertyBasedPMType):
    Id = cp.NodeAttributeProperty('Id')
    text = cp.NodeTextProperty()
    _props = ['Id', 'text']
    def __init__(self, id_, text):
        """
        @param id_: a string
        @param text : a string
        """
        self.Id = id_
        self.text = text
        

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.Id.updateFromNode(obj, node)
        cls.text.updateFromNode(obj, node)
        return obj

    
    
class T_DualChannelDef(PropertyBasedPMType):
    Selector = cp.SubElementListProperty([namespaces.mdpwsTag('Selector')], cls=T_Selector)
    Algorithm = cp.NodeAttributeProperty('Algorithm')
    Transform = cp.NodeAttributeProperty('Transform')
    _props = ['Selector', 'Algorithm', 'Transform']
    
    def __init__(self, selectors, algorithm=None, transform=None):
        """
        @param selectors: a list of Selector objects
        @param algorithm : a string
        @param transform : a string
        """
        self.Selector = selectors
        self.Algorithm = algorithm
        self.Transform = transform
        

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None, None)
        cls.Selector.updateFromNode(obj, node)
        cls.Algorithm.updateFromNode(obj, node)
        cls.Transform.updateFromNode(obj, node)
        return obj



class T_SafetyContextDef(PropertyBasedPMType):
    Selector = cp.SubElementListProperty([namespaces.siTag('Selector')], cls=T_Selector)
    _props = ['Selector',]
    
    def __init__(self, selectors):
        """
        @param selectors: a list of Selector objects
        """
        self.Selector = selectors
        

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        cls.Selector.updateFromNode(obj, node)
        return obj


class T_SafetyReq(PropertyBasedPMType):
    DualChannelDef = cp.SubElementProperty([namespaces.siTag('DualChannelDef')], valueClass=T_DualChannelDef)  # optional
    SafetyContextDef = cp.SubElementProperty([namespaces.siTag('SafetyContextDef')], valueClass=T_SafetyContextDef) #optional
    _props = ['DualChannelDef', 'SafetyContextDef']
    
    def __init__(self, dualChannelDef, safetyContextDef):
        self.DualChannelDef = dualChannelDef
        self.SafetyContextDef = safetyContextDef
        

    @classmethod
    def fromNode(cls, node):
        obj = cls(None, None)
        cls.DualChannelDef.updateFromNode(obj, node)
        cls.SafetyContextDef.updateFromNode(obj, node)
        return obj



class T_Udi(PropertyBasedPMType):
    DeviceIdentifier = cp.NodeTextProperty([namespaces.domTag('DeviceIdentifier')], isOptional=False)
    HumanReadableForm = cp.NodeTextProperty([namespaces.domTag('HumanReadableForm')], isOptional=False)
    Issuer = cp.SubElementProperty([namespaces.domTag('Issuer')], valueClass=InstanceIdentifier)
    Jurisdiction = cp.SubElementProperty([namespaces.domTag('Jurisdiction')], valueClass=InstanceIdentifier)
    _props = ['DeviceIdentifier', 'HumanReadableForm', 'Issuer', 'Jurisdiction']

    def __init__(self, device_identifier=None, humanreadable_form=None, issuer=None, jurisdiction=None):
        """
        UDI fragments as defined by the FDA. (Only used in BICEPS Final)
        :param device_identifier: a string
        :param humanreadable_form: a string
        :param issuer: an InstanceIdentifier
        :param jurisdiction: an InstanceIdentifier (optional)
        """
        self.DeviceIdentifier = device_identifier
        self.HumanReadableForm = humanreadable_form
        self.Issuer = issuer
        self.Jurisdiction = jurisdiction

        
###################################################################################
# following : classes that serve only as name spaces


class StringEnum(str, enum.Enum):

    def __str__(self):
        return self.value


class SafetyClassification(StringEnum):
    INF = 'Inf'
    MED_A = 'MedA'
    MED_B = 'MedB'
    MED_C = 'MedC'


class MdsOperatingMode(StringEnum):
    NORMAL = 'Nml'
    DEMO = 'Dmo'
    SERVICE = 'Srv'
    MAINTENANCE = 'Mtn'
    

class OperatingMode(StringEnum):
    DISABLED = 'Dis'
    ENABLED = 'En'
    NA = 'NA'

    
class ComponentActivation(StringEnum):
    ON = 'On'
    NOT_READY = 'NotRdy'
    STANDBY = 'StndBy'
    OFF = 'Off'
    SHUTDOWN = 'Shtdn'
    FAILURE = 'Fail'


class ContextAssociation(StringEnum):
    NO_ASSOCIATION = 'No'
    PRE_ASSOCIATION = 'Pre'
    ASSOCIATED = 'Assoc'
    DISASSOCIATED = 'Dis'


class AlertConditionMonitoredLimits(StringEnum):
    ALL_ON = 'All'
    LOW_OFF = 'LoOff'
    HIGH_OFF = 'HiOff'
    ALL_OFF = 'None'


class AlertConditionPriority(StringEnum):
    NONE = 'None'
    LOW = 'Lo'
    MEDIUM = 'Me'
    HIGH = 'Hi'


class AlertConditionKind(StringEnum):
    PHYSIOLOGICAL = 'Phy'
    TECHNICAL = 'Tec'
    OTHER = 'Oth'


class AlertActivation(StringEnum):
    ON = 'On'
    OFF = 'Off'
    PAUSED = 'Psd'


class AlertSignalPresence(StringEnum):
    ON = 'On'
    OFF = 'Off'
    LATCH = 'Latch'
    ACK = 'Ack'


class AlertSignalManifestation(StringEnum):
    AUD = 'Aud' # Aud = Audible. The ALERT SIGNAL manifests in an audible manner, i.e., the alert can be heard. Example: an alarm sound.
    VIS = 'Vis' # Vis = Visible. The ALERT SIGNAL manifests in a visible manner, i.e., the alert can be seen. Example: a red flashing light.
    TAN = 'Tan' # Tan = Tangible. The ALERT SIGNAL manifests in a tangible manner, i.e., the alert can be felt. Example: vibration.
    OTH = 'Oth' # Oth = Other. The ALERT SIGNAL manifests in a manner not further specified.


class MetricAvailability(StringEnum):
    INTERMITTENT = 'Intr'
    CONTINUOUS = 'Cont'


class MetricCategory(StringEnum):
    UNSPECIFIED = 'Unspec'
    MEASUREMENT = 'Msrmt'
    CALCULATION = 'Clc'
    SETTING = 'Set'
    PRESETTING = 'Preset'
    RECOMMENDATION = 'Rcmm'


class MeasurementValidity(StringEnum):
    """Level of validity of a measured value.
    Used in BICEPS Final"""
    VALID = 'Vld'
    VALIDATED_DATA = 'Vldated'
    MEASUREMENT_ONGOING = 'Ong'
    QUESTIONABLE = 'Qst'
    CALIBRATION_ONGOING = 'Calib'
    INVALID = 'Inv'
    OVERFLOW = 'Oflw'
    UNDERFLOW = 'Uflw'
    NA = 'NA'

    
class InvocationState(StringEnum): # a namespace class
    WAIT = 'Wait'                  # Wait = Waiting. The operation has been queued and waits for execution. 
    START = 'Start'                # Start = Started. The execution of the operation has been started
    CANCELLED = 'Cnclld'           # Cnclld = Cancelled. The execution has been cancelled by the SERVICE PROVIDER.
    CANCELLED_MANUALLY = 'CnclldMan' # CnclldMan = Cancelled Manually. The execution has been cancelled by the operator.
    FINISHED = 'Fin'               # Fin = Finished. The execution has been finished.
    FINISHED_MOD = 'FinMod'        # FinMod = Finished with modification. As the requested target value could not be reached, the next best value has been chosen and used as target value.
    FAILED = 'Fail'                # The execution has been failed.


class InvocationError(StringEnum):
    UNSPECIFIED = 'Unspec'         # An unspecified error has occurred. No more information about the error is available.
    UNKNOWN_OPERATION = 'Unkn'     # Unknown Operation. The HANDLE to the operation object is not known.
    INVALID_VALUE = 'Inv'          # Invalid Value. The HANDLE to the operation object does not match the invocation request message
    OTHER = 'Oth'                  # Another type of error has occurred. More information on the error MAY be available.


class GenerationMode(StringEnum):
    """Describes whether METRIC data is generated by real measurements or under unreal settings (demo or test data)."""
    REAL = 'Real'                  # Real Data. A value that is generated under real conditions
    TEST = 'Test'                  # Test Data. A value that is arbitrary and is for testing purposes only
    DEMO = 'Demo'                  # Demo Data. A value that is arbitrary and is for demonstration purposes only


class Units(object):
    UnitLess = CodedValue('262656') # used if a metric has no unit


class DescriptionModificationTypes(StringEnum):
    CREATE = 'Crt'
    UPDATE = 'Upt'
    DELETE = 'Del'


class PatientType(StringEnum):
    UNSPECIFIED = 'Unspec'
    ADULT = 'Ad'
    ADOLESCENT = 'Ado'
    PEDIATRIC = 'Ped'
    INFANT = 'Inf'
    NEONATAL = 'Neo'
    OTHER = 'Oth'



from . import pysoap
from . import safety
from . import namespaces
from . import wsdiscovery
from . import mdib
from . import netconn
from . import pmtypes
from . import location
from . import sdcclient
from . import sdcdevice
from . import xmlparsing
from . import commlog
from . import compression

try:
    from . import version
    __version__ = version.version
except ImportError:
    __version__ = '0.0.0'



import threading
import uuid
import copy
import time
import socket
import traceback
import http.client
from http.server import HTTPServer
import queue
import urllib
from lxml import etree as etree_
from sdc11073.pysoap.soapenvelope import Soap12Envelope, ReceivedSoap12Envelope, WsAddress, WsSubscribe, \
    SoapResponseException
from sdc11073.pysoap.soapenvelope import WsaEndpointReferenceType
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import nsmap as _global_nsmap
from ..namespaces import wseTag, wsaTag, WSA_IS_REFERENCE_PARAMETER
from .. import xmlparsing, isoduration
from .. import commlog
from .. import observableproperties as properties
from .. import loghelper
from ..httprequesthandler import HTTPRequestHandler
from sdc11073.pysoap.soapclient import HTTPReturnCodeError

MULTITHREADED = True
SUBSCRIPTION_CHECK_INTERVAL = 5  # seconds


class MyThreadingMixIn(object):

    def process_request_thread(self, request, client_address):
        """Same as in BaseServer but as a thread.

        In addition, exception handling is done here.

        """
        try:
            self.finish_request(request, client_address)
            self.shutdown_request(request)
        except Exception as ex:
            if self.dispatcher is not None:
                # only
                self.handle_error(request, client_address)
            self.shutdown_request(request)

    def process_request(self, request, client_address):
        """Start a new thread to process the request."""
        t = threading.Thread(target=self.process_request_thread,
                             args=(request, client_address),
                             name='SubscrRecv{}'.format(client_address))
        t.daemon = True
        t.start()
        self.threads.append((t, request, client_address))


if MULTITHREADED:
    class MyHTTPServer(MyThreadingMixIn, HTTPServer):
        """ Each request is handled in a thread.
        Following receipe from https://pymotw.com/2/BaseHTTPServer/index.html#module-BaseHTTPServer
        """

        def __init__(self, *args, **kwargs):
            HTTPServer.__init__(self, *args, **kwargs)
            self.daemon_threads = True
            self.threads = []
            self.dispatcher = None

else:
    MyHTTPServer = HTTPServer  # single threaded, sequential operation


class ClSubscription(object):
    """ This class handles a subscription to an event source.
    It stores all key data of the subscription and can renew and unsubscribe this subscription."""
    notification = properties.ObservableProperty()
    IDENT_TAG = etree_.QName('http.local.com', 'MyClIdentifier')

    def __init__(self, dpwsHosted, actions, notification_url, endTo_url, ident, xml_validator):
        """
        @param serviceClient:
        @param filter_:
        @param notification_url: e.g. http://1.2.3.4:9999, or https://1.2.3.4:9999
        """
        self.dpwsHosted = dpwsHosted
        self._actions = actions
        self._filter = ' '.join(actions)
        self._notification_url = notification_url
        self.isSubscribed = False
        self.end_status = None  # if device sent a SubscriptionEnd message, this contains the status from the message
        self.end_reason = None  # if device sent a SubscriptionEnd message, this contains the reason from the message
        self.expireAt = None
        self.expire_minutes = None
        self.dev_reference_param = None
        self.notifyTo_identifier = etree_.Element(self.IDENT_TAG)
        self.notifyTo_identifier.text = uuid.uuid4().urn

        self._endTo_url = endTo_url
        self._endTo_identifier = etree_.Element(self.IDENT_TAG)
        self._endTo_identifier.text = uuid.uuid4().urn

        self._subscriptionManagerAddress = None
        self._logger = loghelper.getLoggerAdapter('sdc.client.subscr', ident)
        self.eventCounter = 0  # for display purpose, we count notifications
        self.cl_ident = ident
        self._device_epr = urllib.parse.urlparse(self.dpwsHosted.endpointReferences[0].address).path
        self._xml_validator = xml_validator

    @property
    def end_to_identifier(self):
        return self._endTo_identifier

    def _mkSubscribeEnvelope(self, subscribe_epr, expire_minutes):
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/Subscribe',
                                          to=subscribe_epr))

        body = WsSubscribe(notifyTo=WsaEndpointReferenceType(self._notification_url,
                                                             referenceParametersNode=[self.notifyTo_identifier]),
                           endTo=WsaEndpointReferenceType(self._endTo_url,
                                                          referenceParametersNode=[self._endTo_identifier]),
                           expires=expire_minutes * 60,
                           filter_=self._filter)
        soapEnvelope.addBodyObject(body)
        return soapEnvelope

    def _handleSubscribeResponse(self, soapEnvelope):
        # Check content of response; raise Error if subscription was not successful
        try:
            msgNode = soapEnvelope.msgNode
            if msgNode.tag == wseTag('SubscribeResponse'):
                address = msgNode.xpath('wse:SubscriptionManager/wsa:Address/text()', namespaces=_global_nsmap)
                self.dev_reference_param = None

                reference_params = msgNode.xpath('wse:SubscriptionManager/wsa:ReferenceParameters',
                                                 namespaces=_global_nsmap)
                if reference_params:
                    self.dev_reference_param = reference_params[0]
                expires = msgNode.xpath('wse:Expires/text()', namespaces=_global_nsmap)

                self._subscriptionManagerAddress = urllib.parse.urlparse(address[0])
                expireseconds = isoduration.parse_duration(expires[0])
                self.expireAt = time.time() + expireseconds
                self.isSubscribed = True
                self._logger.info('Subscribe was successful: expires at {}, address="{}"',
                                  self.expireAt, self._subscriptionManagerAddress)
            else:
                # This is a failure response or even rubbish. log it and raise error
                self._logger.error('Subscribe response has unexpected content: {}', soapEnvelope.as_xml(pretty=True))
                self.isSubscribed = False
                raise SoapResponseException(soapEnvelope)
        except AttributeError:
            self._logger.error('Subscribe response has unexpected content: {}', soapEnvelope.as_xml(pretty=True))
            self.isSubscribed = False
            raise SoapResponseException(soapEnvelope)

    def subscribe(self, expire_minutes=60):
        self._logger.info('### startSubscription "{}" ###', self._filter)
        self.eventCounter = 0
        self.expire_minutes = expire_minutes  # saved for later renewal, we will use the same interval
        # ToDo: check if there is more than one address. In that case a clever selection is needed
        address = self.dpwsHosted.endpointReferences[0].address
        soapEnvelope = self._mkSubscribeEnvelope(address, expire_minutes)
        msg = 'subscribe {}'.format(self._filter)
        try:
            resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._device_epr, soapEnvelope, msg=msg)
            self._handleSubscribeResponse(resultSoapEnvelope)
        except HTTPReturnCodeError as ex:
            self._logger.error('could not subscribe: {}'.format(HTTPReturnCodeError))

    def _add_device_references(self, soapEnvelope):
        """ add references for requests to device (renew, getstatus, unsubscribe)"""
        if self.dev_reference_param is not None:
            for e in self.dev_reference_param:
                e_ = copy.copy(e)
                # mandatory attribute acc. to ws_addressing SOAP Binding (https://www.w3.org/TR/2006/REC-ws-addr-soap-20060509/)
                e_.set(wsaTag('IsReferenceParameter'), 'true')
                soapEnvelope.addHeaderElement(e_)

    def _mkRenewEnvelope(self, expire_minutes):
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/Renew',
                                          to=urllib.parse.urlunparse(self._subscriptionManagerAddress)))
        self._add_device_references(soapEnvelope)
        renewNode = etree_.Element(wseTag('Renew'), nsmap=Prefix.partialMap(Prefix.WSE))
        expiresNode = etree_.SubElement(renewNode, wseTag('Expires'), nsmap=Prefix.partialMap(Prefix.WSE))
        expiresNode.text = isoduration.durationString(expire_minutes * 60)
        soapEnvelope.addBodyElement(renewNode)
        return soapEnvelope

    def _handleRenewResponse(self, soapEnvelope):
        # Check content of response; raise Error if subscription was not successful
        bodyNode = soapEnvelope.bodyNode
        renewResponse = bodyNode.xpath('wse:RenewResponse', namespaces=_global_nsmap)
        if len(renewResponse) == 1:
            # this means renew was accepted
            expires = bodyNode.xpath('wse:RenewResponse/wse:Expires/text()', namespaces=_global_nsmap)
            expireseconds = isoduration.parse_duration(expires[0])
            self.expireAt = time.time() + expireseconds
        else:
            raise SoapResponseException(soapEnvelope)

    def renew(self, expire_minutes=60):
        soapEnvelope = self._mkRenewEnvelope(expire_minutes)
        try:
            resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._subscriptionManagerAddress.path,
                                                                               soapEnvelope, msg='renew')
            self._logger.debug('{}', resultSoapEnvelope.as_xml(pretty=True))
        except HTTPReturnCodeError as ex:
            self.isSubscribed = False
            self._logger.error('could not renew: {}'.format(HTTPReturnCodeError))
        except (http.client.HTTPException, ConnectionError) as ex:
            self._logger.warn('renew failed: {}', ex)
            self.isSubscribed = False
        except Exception as ex:
            self._logger.error('Exception in renew: {}', ex)
            self.isSubscribed = False
        else:
            try:
                self._handleRenewResponse(resultSoapEnvelope)
                return self.remainingSubscriptionSeconds
            except SoapResponseException as ex:
                self.isSubscribed = False
                self._logger.warn('renew failed: {}',
                                  etree_.tostring(ex.soapResponseEnvelope.bodyNode, pretty_print=True))

    def unsubscribe(self):
        if not self.isSubscribed:
            return

        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/Unsubscribe',
                                          to=urllib.parse.urlunparse(self._subscriptionManagerAddress)))
        self._add_device_references(soapEnvelope)
        soapEnvelope.addBodyElement(etree_.Element(wseTag('Unsubscribe')))
        resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._subscriptionManagerAddress.path,
                                                                           soapEnvelope, msg='unsubscribe')
        responseAction = resultSoapEnvelope.address.action
        # check response: response does not contain explicit status. If action== UnsubscribeResponse all is fine.
        if responseAction == 'http://schemas.xmlsoap.org/ws/2004/08/eventing/UnsubscribeResponse':
            self._logger.info('unsubscribe: end of subscription {} was confirmed.', self._filter)
        else:
            self._logger.error('unsubscribe: unexpected response action: {}', resultSoapEnvelope.as_xml(pretty=True))
            raise RuntimeError(
                'unsubscribe: unexpected response action: {}'.format(resultSoapEnvelope.as_xml(pretty=True)))

    def _mkGetStatusEnvelope(self):
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/GetStatus',
                                          to=urllib.parse.urlunparse(self._subscriptionManagerAddress)))
        self._add_device_references(soapEnvelope)
        bodyNode = etree_.Element(wseTag('GetStatus'))
        soapEnvelope.addBodyElement(bodyNode)
        return soapEnvelope

    def getStatus(self):
        """ Sends a GetStatus Request to the device.
        @return: the remaining time of the subscription or None, if the request was not successful
        """
        soapEnvelope = self._mkGetStatusEnvelope()
        try:
            resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._subscriptionManagerAddress.path,
                                                                               soapEnvelope, msg='getStatus')
        except HTTPReturnCodeError as ex:
            self.isSubscribed = False
            self._logger.error('could not get status: {}'.format(HTTPReturnCodeError))
        except (http.client.HTTPException, ConnectionError) as ex:
            self.isSubscribed = False
            self._logger.warn('getStatus: Connection Error {} for subscription {}', ex, self._filter)
        except Exception as ex:
            self._logger.error('Exception in getStatus: {}', ex)
            self.isSubscribed = False
        else:
            try:
                expiresNode = resultSoapEnvelope.msgNode.find('wse:Expires', namespaces=_global_nsmap)
                if expiresNode is None:
                    self._logger.warn('getStatus for {}: Could not find "Expires" node! getStatus={} ', self._filter,
                                      resultSoapEnvelope.rawdata)
                    raise SoapResponseException(resultSoapEnvelope)
                else:
                    expires = expiresNode.text
                    expiresValue = isoduration.parse_duration(expires)
                    self._logger.debug('getStatus for {}: Expires = {} = {} seconds, counter = {}', self._filter,
                                       expires,
                                       expiresValue,
                                       self.eventCounter)
                    return expiresValue
            except AttributeError:
                self._logger.warn('No msg in envelope')

    def checkStatus(self, renewLimit):
        """ Calls getStatus and updates internal data.
        @param renewLimit: a value in seconds. If remaining duration of subscription is less than this value, it renews the subscription.
        @return: None
        """
        if not self.isSubscribed:
            return

        remainingTime = self.getStatus()
        if remainingTime is None:
            self.isSubscribed = False
            return
        elif abs(remainingTime - self.remainingSubscriptionSeconds) > 10:
            self._logger.warn(
                'time delta between expected expire and reported expire  > 10 seconds. Will correct own expectation.')
            self.expireAt = time.time() + remainingTime

        if self.remainingSubscriptionSeconds < renewLimit:
            self._logger.info('renewing subscription')
            self.renew()

    def checkStatus_renew(self):
        """ Calls renew and updates internal data.
        @return: None
        """
        if self.isSubscribed:
            self.renew()

    @property
    def remainingSubscriptionSeconds(self):
        return self.expireAt - time.time()

    def onNotification(self, soapEnvelope):
        try:
            soapEnvelope.validate_envelope(self._xml_validator)
        except etree_.DocumentInvalid as ex:
            self._logger.error('received invalid document: {}', ex)
        except Exception as ex:
            self._logger.error('error validation document: {}', ex)
        self.eventCounter += 1
        self.notification = soapEnvelope

    @property
    def shortFilterString(self):
        return xmlparsing.shortFilterString(self._actions)

    def __str__(self):
        return 'Subscription of "{}", isSubscribed={}, remaining time = {} sec., count={}'.format(
            self.shortFilterString,
            self.isSubscribed,
            int(self.remainingSubscriptionSeconds),
            self.eventCounter)


class SubscriptionManager(threading.Thread):
    """ Factory for Subscription objects, thread that automatically renews expiring subscriptions.
    @param notification_url: the destination url for notifications.
    @param endTo_url: if given the destination url for end subscription notifications; if not given, the notification_url is used.
    @param check_interval: the interval (in seconds ) for getStatus requests. Defaults to SUBSCRIPTION_CHECK_INTERVAL
    @param ident: a string that is used in log output; defaults to empty string
     """
    allSubscriptionsOkay = properties.ObservableProperty(True)  # a boolean
    keepAlive_with_renew = True  # enable as workaround if checkstatus is not supported

    def __init__(self, notification_url, endTo_url=None, checkInterval=None, log_prefix='', xml_validator=None):
        super().__init__(name='Cl_SubscriptionManager{}'.format(log_prefix))
        self.daemon = True
        self._checkInterval = checkInterval or SUBSCRIPTION_CHECK_INTERVAL
        self.subscriptions = {}
        self._subscriptionsLock = threading.Lock()

        self._run = False
        self._notification_url = notification_url
        self._endTo_url = endTo_url or notification_url
        self._logger = loghelper.getLoggerAdapter('sdc.client.subscrMgr', log_prefix)
        self.log_prefix = log_prefix
        self._xml_validator = xml_validator

    def stop(self):
        self._run = False
        self.join(timeout=2)
        with self._subscriptionsLock:
            self.subscriptions.clear()

    def run(self):
        self._run = True
        try:
            while self._run:
                try:
                    for i in range(self._checkInterval):
                        time.sleep(1)
                        if not self._run:
                            return
                        # check if all subscriptions are okay
                        with self._subscriptionsLock:
                            not_okay = [s for s in self.subscriptions.values() if not s.isSubscribed]
                            subscribed = [s.isSubscribed for s in self.subscriptions.values()]
                            self.allSubscriptionsOkay = (len(not_okay) == 0)
                    with self._subscriptionsLock:
                        subscriptions = list(self.subscriptions.values())
                    for subscription in subscriptions:
                        if self.keepAlive_with_renew:
                            subscription.checkStatus_renew()
                        else:
                            subscription.checkStatus(renewLimit=self._checkInterval * 5)
                    self._logger.debug('##### SubscriptionManager Interval ######')
                    for subscription in subscriptions:
                        self._logger.debug('{}', subscription)
                except Exception as ex:
                    self._logger.error('##### check loop: {}', traceback.format_exc())
        finally:
            self._logger.info('terminating subscriptions check loop! self._run={}', self._run)

    def mkSubscription(self, dpwsHosted, filters):
        s = ClSubscription(dpwsHosted, filters, self._notification_url, self._endTo_url, self.log_prefix,
                           self._xml_validator)
        filter_ = ' '.join(filters)
        with self._subscriptionsLock:
            self.subscriptions[filter_] = s
        return s

    def onSubScriptionEnd(self, soapenvelope):
        subscr_ident_list = soapenvelope.headerNode.findall(ClSubscription.IDENT_TAG, namespaces=_global_nsmap)
        statuus = soapenvelope.bodyNode.xpath('wse:SubscriptionEnd/wse:Status/text()', namespaces=_global_nsmap)
        reasons = soapenvelope.bodyNode.xpath('wse:SubscriptionEnd/wse:Reason/text()', namespaces=_global_nsmap)
        if statuus:
            info = ' status={} '.format(statuus[0])
        else:
            info = ''
        if reasons:
            if len(reasons) == 1:
                info += ' reason = {}'.format(reasons[0])
            else:
                info += ' reasons = {}'.format(reasons)
        if not subscr_ident_list:
            self._logger.warn('onSubScriptionEnd: did not find any identifier in message')
            return None
        subscr_ident = subscr_ident_list[0]
        for s in self.subscriptions.values():
            if subscr_ident.text == s.end_to_identifier.text:
                self._logger.info('onSubScriptionEnd: received Subscription End for {} {}',
                                  s.shortFilterString,
                                  info)
                s.isSubscribed = False
                if len(statuus) > 0:
                    s.end_status = statuus[0]
                if len(reasons) > 0:
                    s.end_reason = reasons[0]
                return s
        self._logger.warn('onSubScriptionEnd: have no subscription for identifier = {}', subscr_ident.text)
        return None

    def unsubscribeAll(self):
        with self._subscriptionsLock:
            current_subscriptions = list(self.subscriptions.values())  # make a copy
            self.subscriptions.clear()
            for s in current_subscriptions:
                try:
                    s.unsubscribe()
                except:
                    self._logger.warn('unsubscribe error: {}\n call stack:{} ', traceback.format_exc(),
                                      traceback.format_stack())


class _DispatchError(Exception):
    def __init__(self, httpErrorcode, errorText):
        super(_DispatchError, self).__init__()
        self.httpErrorcode = httpErrorcode
        self.errorText = errorText


class SOAPNotificationsDispatcher(object):
    """ receiver of all notifications"""

    def __init__(self, log_prefix, sdc_definitions):
        self._logger = loghelper.getLoggerAdapter('sdc.client.notif_dispatch', log_prefix)
        self.log_prefix = log_prefix
        self._sdc_definitions = sdc_definitions
        self.methods = {}

    def register_function(self, action, ref, fn):
        self.methods[(ref, action)] = fn

    def dispatch(self, path, xml):
        start = time.time()

        fn, action, request = self._dispatch_process_data(xml, path)
        fn(request)

        duration = time.time() - start
        if duration > 0.005:
            self._logger.debug('action {}: duration = {:.4f}sec', action, duration)
        return ''

    def _dispatch_process_data(self, xml, path):
        normalized_xml = self._sdc_definitions.normalizeXMLText(xml)
        request = ReceivedSoap12Envelope.fromXMLString(normalized_xml)
        try:
            action = request.address.action
        except AttributeError:
            raise _DispatchError(404, 'no action in request')
        self._logger.debug('received notification path={}, action = {}', path, action)

        ref_parameters = request.headerNode.findall(ClSubscription.IDENT_TAG, namespaces=_global_nsmap)
        ref_parameters = [i for i in ref_parameters if i.get(WSA_IS_REFERENCE_PARAMETER) in ("1", "true")]

        if len(ref_parameters) != 1:
            raise _DispatchError(400, f'Expected exactly one matching reference parameter missing from request.')

        try:
            fn = self.methods[(ref_parameters[0].text, action)]
        except KeyError:
            self._logger.error('action "{}" not registered. Known:{}'.format(action, self.methods.keys()))
            raise _DispatchError(404, 'action not registered')

        return fn, action, request


class SOAPNotificationsDispatcherThreaded(SOAPNotificationsDispatcher):

    def __init__(self, ident, bicepsSchema):
        super(SOAPNotificationsDispatcherThreaded, self).__init__(ident, bicepsSchema)
        self._queue = queue.Queue(1000)
        self._worker = threading.Thread(target=self._readqueue)
        self._worker.daemon = True
        self._worker.start()

    def dispatch(self, path, xml):
        fn, action, request = self._dispatch_process_data(xml, path)
        self._queue.put((fn, request, action))
        return ''

    def _readqueue(self):
        while True:
            fn, request, action = self._queue.get()
            try:
                fn(request)
            except:
                self._logger.error(
                    'method {} for action "{}" failed:{}'.format(fn.__name__, action, traceback.format_exc()))


class SOAPNotificationsHandler(HTTPRequestHandler):
    disable_nagle_algorithm = True
    wbufsize = 0xffff  # 64k buffer to prevent tiny packages
    RESPONSE_COMPRESS_MINSIZE = 256  # bytes, compress response it it is larger than this value (and other side supports compression)

    def do_POST(self):
        """SOAP POST gateway"""
        self.server.threadObj._logger.debug('notification do_POST incoming')  # pylint: disable=protected-access
        dispatcher = self.server.dispatcher
        response_string = ''
        if dispatcher is None:
            # close this connection
            self.close_connection = 1
            self.server.threadObj._logger.warn(
                'received a POST request, but no dispatcher => returning 404 ')  # pylint:disable=protected-access
            self.send_response(404)  # not found
        else:
            request_bytes = self._read_request()

            self.server.threadObj._logger.debug('notification {} bytes',
                                                request_bytes)  # pylint: disable=protected-access
            # execute the method
            commlog.defaultLogger.logSoapSubscrMsgIn(request_bytes)
            try:
                response_string = self.server.dispatcher.dispatch(self.path, request_bytes)
                if response_string is None:
                    response_string = ''
                self.send_response(202, b'Accepted')
            except _DispatchError as ex:
                self.server.threadObj._logger.error('received a POST request, but got _DispatchError => returning {}',
                                                    ex.httpErrorcode)  # pylint:disable=protected-access
                self.send_response(ex.httpErrorcode, ex.errorText)
            except Exception as ex:
                self.server.threadObj._logger.error(
                    'received a POST request, but got Exception "{}"=> returning {}\n{}', ex, 500,
                    traceback.format_exc())  # pylint:disable=protected-access
                self.send_response(500, b'server error in dispatch')
        response_bytes = response_string.encode('utf-8')
        if len(response_bytes) > self.RESPONSE_COMPRESS_MINSIZE:
            response_bytes = self._compressIfRequired(response_bytes)

        self.send_header("Content-Type", "application/soap+xml; charset=utf-8")
        self.send_header("Content-Length", len(response_bytes))  # this is necessary for correct keep-alive handling!
        self.end_headers()
        self.wfile.write(response_bytes)


class NotificationsReceiverDispatcherThread(threading.Thread):

    def __init__(self, my_ipaddress, sslContext, log_prefix, sdc_definitions, supportedEncodings,
                 soap_notifications_handler_class=None, async_dispatch=True):
        """

        :param my_ipaddress: http server will listen on this address
        :param sslContext: http server uses this ssl context
        :param ident: used for logging
        :param sdc_definitions: namespaces etc
        :param supportedEncodings: a list of strings
        :param soap_notifications_handler_class: if None, SOAPNotificationsHandler is used,
                otherwise the provided class ( a HTTPRequestHandler).
        :param async_dispatch: if True, incoming requests are queued and response is sent (processing is done later).
                                if False, response is sent after the complete processing is done.
        """
        super(NotificationsReceiverDispatcherThread, self).__init__(
            name='Cl_NotificationsReceiver_{}'.format(log_prefix))
        self._sslContext = sslContext
        self._soap_notifications_handler_class = soap_notifications_handler_class
        self.daemon = True
        self._logger = loghelper.getLoggerAdapter('sdc.client.notif_dispatch', log_prefix)

        self._my_ipaddress = my_ipaddress
        self.my_port = None
        self.base_url = None
        self.httpd = None
        self.supportedEncodings = supportedEncodings
        # create and set up the dispatcher for notifications
        if async_dispatch:
            self.dispatcher = SOAPNotificationsDispatcherThreaded(log_prefix, sdc_definitions)
        else:
            self.dispatcher = SOAPNotificationsDispatcher(log_prefix, sdc_definitions)
        self.started_evt = threading.Event()  # helps to wait until thread has initialised is variables

    def run(self):
        try:
            myport = 0  # zero means that OS selects a free port
            self.httpd = MyHTTPServer((self._my_ipaddress, myport),
                                      self._soap_notifications_handler_class or SOAPNotificationsHandler)
            # add use compression flag to the server
            setattr(self.httpd, 'supportedEncodings', self.supportedEncodings)
            self.my_port = self.httpd.server_port
            self._logger.info('starting Notification receiver on {}:{}', self._my_ipaddress, self.my_port)
            if self._sslContext:
                self.httpd.socket = self._sslContext.wrap_socket(self.httpd.socket)
                self.base_url = 'https://{}:{}/'.format(self._my_ipaddress, self.my_port)
            else:
                self.base_url = 'http://{}:{}/'.format(self._my_ipaddress, self.my_port)
            self.httpd.dispatcher = self.dispatcher
            self.httpd.threadObj = self  # make logger available for SOAPNotificationsHandler
            self.started_evt.set()
            self.httpd.serve_forever()
        except Exception:
            self._logger.error(
                'Unhandled Exception at thread runtime. Thread will abort! {}'.format(traceback.format_exc()))
            raise

    def stop(self, closeAllConnections=True):
        """
        @param closeAllConnections: for testing purpose one might want to keep the connection handler threads alive.
                If param is False then they are kept alive.
        """
        self.httpd.shutdown()
        self.httpd.socket.close()
        if closeAllConnections:
            if self.httpd.dispatcher is not None:
                self.httpd.dispatcher.methods = {}
                self.httpd.dispatcher = None  # this leads to a '503' reaction in SOAPNotificationsHandler
            for thr in self.httpd.threads:
                thread, request, client_addr = thr
                if thread.is_alive():
                    try:
                        request.shutdown(socket.SHUT_RDWR)
                        request.close()
                        self._logger.info('closed socket for notifications from {}', client_addr)
                    except OSError as ex:
                        # the connection is already closed
                        continue
                    except Exception as ex:
                        self._logger.warn('error closing socket for notifications from {}: {}', client_addr, ex)
            time.sleep(0.1)
            for thr in self.httpd.threads:
                thread, request, client_addr = thr
                if thread.is_alive():
                    thread.join(2)
            del self.httpd.threads[:]



import time
import random
import traceback
from threading import Thread, Event
from .. import namespaces
from ..pmtypes import  AlertActivation, AlertConditionKind, AlertSignalPresence, SystemSignalActivation, AlertSignalManifestation
from . import providerbase


class GenericAlarmProvider(providerbase.ProviderRole):
    WORKERTHREAD_INTERVAL = 1.0 # seconds
    def __init__(self, log_prefix):
        super(GenericAlarmProvider, self).__init__(log_prefix)

        # some time stamps for handling of delegable alert signals
        self._lastActivateAllDelegableAlerts = 0 # time when _activateAllDelegableAlertSignals has been called last time
        self._lastSetAlertSignalState = {} # a lookup by alert signal handle , value = time of last call

        self._stopWorker = Event()
        self._workerThread = None

    def initOperations(self, mdib):
        super(GenericAlarmProvider, self).initOperations(mdib)
        self._setAlertSystemStatesInitialValues()
        self._setAlertStatesInitialValues()
        self._workerThread = Thread(target=self._workerThreadLoop)
        self._workerThread.daemon = True
        self._workerThread.start()

    def makeMissingOperations(self):
        return []

    def stop(self):
        self._stopWorker.set()
        self._workerThread.join()


    def makeOperationInstance(self, operationDescriptorContainer):
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        operationTargetDescr = self._mdib.descriptions.handle.getOne(operationTargetHandle) # descriptor container
        if operationDescriptorContainer.NODETYPE == namespaces.domTag('SetValueOperationDescriptor'):
            pass
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('ActivateOperationDescriptor'):
            pass
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('SetAlertStateOperationDescriptor'):
            if operationTargetDescr.NODETYPE == namespaces.domTag('AlertSignalDescriptor'):
                # no check for code, because the setAlertState operation always means setting
                # ActivationState, Presence and ActualSignalGenerationDelay
                # if stricter checking needed, one might add it
                operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                     currentArgumentHandler=self._setAlertSignalState)

                self._logger.info(
                    'GenericAlarmProvider: added handler "self._setAlertState" for {} target= {} '.format(operationDescriptorContainer,
                                                                                                          operationTargetDescr))
                return operation

        return None  # None == no handler for this operation instantiated

    def _setAlertSystemStatesInitialValues(self):
        alertSystemStates = self._mdib.states.NODETYPE.get(namespaces.domTag('AlertSystemState'), [])
        for alertSystemState in alertSystemStates:
            if hasattr(alertSystemState, 'SystemSignalActivation'): # attribute not exists in Draft6
                alertSystemState.SystemSignalActivation.append(SystemSignalActivation(manifestation=AlertSignalManifestation.AUD,
                                                                                              state=AlertActivation.ON))

    def _setAlertStatesInitialValues(self):
        '''
        - if an AlertCondition.ActivationState is 'On', then the local AlertSignals shall also be 'On'
        - all remote alert Signals shall be 'Off' initially (must be explicitely enabled by delegating device)'''
        for alertCondition in self._mdib.states.NODETYPE.get(namespaces.domTag('AlertConditionState'),[]):
            alertCondition.ActivationState = AlertActivation.ON
        for alertCondition in self._mdib.states.NODETYPE.get(namespaces.domTag('LimitAlertConditionState'),[]):
            alertCondition.ActivationState = AlertActivation.ON
        for alertSignalState in self._mdib.states.NODETYPE.get(namespaces.domTag('AlertSignalState'),[]):
            if alertSignalState.Location == 'Rem':
                alertSignalState.ActivationState = AlertActivation.OFF
                alertSignalState.updateNode()
            else:
                alertSignalDescr = self._mdib.descriptions.handle.getOne(alertSignalState.descriptorHandle)
                # ConditionSignaled can be None, in that case do nothing
                if alertSignalDescr.ConditionSignaled:
                    alertConditionState = self._mdib.states.descriptorHandle.getOne(alertSignalDescr.ConditionSignaled, allowNone=True)
                    if alertConditionState and alertSignalState.ActivationState != alertConditionState.ActivationState:
                        alertSignalState.ActivationState = alertConditionState.ActivationState
                        alertSignalState.updateNode()

    def _getDescriptor(self, handle, mdib, transaction):
        ''' Helper that looks for descriptor first in current transaction, then in mdib. returns first found one or raises KeyError'''
        descriptor = None
        tr_item = transaction.descriptorUpdates.get(handle)
        if tr_item is not None:
            descriptor = tr_item.new
        if descriptor is None:
            # it is not part of this transaction
            descriptor = mdib.descriptions.handle.getOne(handle, allowNone=True)
        if descriptor is None:
            raise KeyError('there is no descriptor for {}'.format(handle))
        return descriptor

    def _getChangedAlertConditionStates(self, transaction):
        result = []
        for item in list(transaction.alertStateUpdates.values()):
            tmp = item.old if item.new is None else item.new
            if tmp.NODETYPE in(namespaces.domTag('AlertConditionState'),
                               namespaces.domTag('LimitAlertConditionState')):
                result.append(tmp)
        return result

    def _findAlertSystemsWithModifications(self, mdib, transaction):
        # find all alert systems with changed states
        alertSystemStates = set()
        changedAlertConditions = self._getChangedAlertConditionStates(transaction)
        for tmp in changedAlertConditions:
            alertDescriptor = self._getDescriptor(tmp.descriptorHandle, mdib, transaction)
            alertSystemDescriptor = self._getDescriptor(alertDescriptor.parentHandle, mdib, transaction)
            if alertSystemDescriptor.handle in transaction.alertStateUpdates:
                tmp_st = transaction.alertStateUpdates[alertSystemDescriptor.handle]
                if tmp_st.new is not None:
                    alertSystemStateContainer = tmp_st.new
                    alertSystemStates.add(alertSystemStateContainer)
            else:
                alertSystemStateContainer = transaction.getAlertState(alertSystemDescriptor.handle)
                alertSystemStates.add(alertSystemStateContainer)
        return alertSystemStates

    def onPreCommit(self, mdib, transaction):
        if not transaction.alertStateUpdates:
            return
        # find all alert systems with changed states
        alertSystemStates = self._findAlertSystemsWithModifications( mdib, transaction)
        if alertSystemStates:
            self._updateAlertSystemStates(mdib, transaction, alertSystemStates) # add found alert system states to transaction

        # change AlertSignal Settings in order to be compliant with changed Alert Conditions
        changedAlertConditions = self._getChangedAlertConditionStates(transaction)
        for changedAlertCondition in changedAlertConditions:
            self._updateAlertSignals(changedAlertCondition, mdib, transaction)

    def _updateAlertSystemStates(self, mdib, transaction, alertSystemStates):
        '''update alert system states and add them to transaction
        '''
        def _getAlertState(descriptorHandle):
            alertState = None
            tr_item = transaction.getStateTransactionItem(descriptorHandle)
            if tr_item is not None:
                alertState = tr_item.new
            if alertState is None:
                # it is not part of this transaction
                alertState = mdib.states.descriptorHandle.getOne(descriptorHandle, allowNone=True)
            if alertState is None:
                raise RuntimeError('there is no alert state for {}'.format(descriptorHandle))
            return alertState
        for st in alertSystemStates:
            all_child_descriptors = mdib.descriptions.parentHandle.get(st.descriptorHandle, list())
            all_child_descriptors.extend([i.new for i in transaction.descriptorUpdates.values() if i.new.parentHandle == st.descriptorHandle])
            all_alert_condition_descr = [d for d in all_child_descriptors if hasattr(d, 'Kind')]
            # select all state containers with technical alarms present
            all_tech_descr = [ d for d in all_alert_condition_descr if d.Kind == AlertConditionKind.TECHNICAL]
            all_tech_states = [_getAlertState(d.handle) for d in all_tech_descr]
            all_tech_states = [s for s in all_tech_states if s is not None]
            all_present_tech_states = [s for s in all_tech_states if s.Presence]
            # select all state containers with physiolocical alarms present
            all_phys_descr = [ d for d in all_alert_condition_descr if d.Kind == AlertConditionKind.PHYSIOLOGICAL]
            all_phys_states = [_getAlertState(d.handle) for d in all_phys_descr]
            all_phys_states = [ s for s in all_phys_states if s is not None]
            all_present_phys_states = [s for s in all_phys_states if s.Presence]

            st.PresentTechnicalAlarmConditions = [s.descriptorHandle for s in all_present_tech_states]
            st.PresentPhysiologicalAlarmConditions = [s.descriptorHandle for s in all_present_phys_states]

            st.LastSelfCheck = time.time()
            st.SelfCheckCount = 1 if st.SelfCheckCount is None else st.SelfCheckCount + 1

    def _updateAlertSignals(self, changedAlertCondition, mdib, transaction):
        ''' Handle alert signals for a changed alert condition.
        This method only changes states of local signals.
        Handling of delegated signals is in the responsibility of the delegated device!'''
        alertSignalDescriptors = mdib.descriptions.ConditionSignaled.get(changedAlertCondition.descriptorHandle, [])
        # separate remote from local
        remoteAlertSignalDescriptors = [a for a in alertSignalDescriptors if a.SignalDelegationSupported]
        localAlertSignalDescriptors = [a for a in alertSignalDescriptors if not a.SignalDelegationSupported]

        # look for active delegations (we only need the Manifestation value here)
        active_delegate_manifestations = []
        for a in remoteAlertSignalDescriptors:
            alertSignalState = mdib.states.descriptorHandle.getOne(a.handle)
            if alertSignalState.Presence != AlertSignalPresence.OFF and alertSignalState.Location == 'Rem':
                active_delegate_manifestations.append(a.Manifestation)

        # this lookup gives the values that a local signal shall have:
        # key = (Cond.Presence, isDelegated): value = (SignalState.ActivationState, SignalState.Presence)
        # see BICEPS standard table 9: valid combinations of alert activation states, alert condition presence, ...
        # this is the relevant subset for our case
        lookup = {(True, True): (AlertActivation.PAUSED, AlertSignalPresence.OFF),
                  (True, False): (AlertActivation.ON, AlertSignalPresence.ON),
                  (False, True): (AlertActivation.PAUSED, AlertSignalPresence.OFF),
                  (False, False): (AlertActivation.ON, AlertSignalPresence.OFF)
                  }
        for a in localAlertSignalDescriptors:
            tr_item = transaction.getStateTransactionItem(a.handle)
            if tr_item is None:
                isDelegated = a.Manifestation in active_delegate_manifestations  # is this local signal delegated?
                activation, presence = lookup[(changedAlertCondition.Presence, isDelegated)]
                alertSignalState = transaction.getAlertState(a.handle)

                if alertSignalState.ActivationState != activation or alertSignalState.Presence != presence:
                    alertSignalState.ActivationState = activation
                    alertSignalState.Presence = presence
                else:
                    # don't change
                    transaction.ungetState(alertSignalState)

    def _setUpperLimit(self, operationDescriptorContainer, value):
        ''' set upper limit of an LimitAlertConditionStateContainer'''
        operationTargetHandle = operationDescriptorContainer.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            operationStateContainer = mgr._deviceMdibContainer.states.descriptorHandle.getOne(operationDescriptorContainer.handle)
            self._inAllowedRange(operationStateContainer, value)
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('set upper limit handle="{}" from {} to {}', operationTargetHandle, state.Limits.Upper, value)
            state.Limits.Upper = value

    def _setLowerLimit(self, operationDescriptorContainer, value):
        ''' set lower limit of an LimitAlertConditionStateContainer'''
        operationTargetHandle = operationDescriptorContainer.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            operationStateContainer = mgr._deviceMdibContainer.states.descriptorHandle.getOne(operationDescriptorContainer.handle)
            self._inAllowedRange(operationStateContainer, value)
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('set lower limit handle={} from {} to {}', operationTargetHandle, state.Limits.Lower, value)
            state.Limits.Lower = value

    @staticmethod
    def _inAllowedRange(operationStateContainer, value):
        if operationStateContainer.AllowedRange:
            max_upper = max([i.Upper for i in operationStateContainer.AllowedRange])
            min_lower = min([i.Lower for i in operationStateContainer.AllowedRange])
            if not min_lower <= value <= max_upper:
                raise ValueError('value "{}" to be set is not in AllowedRanges {}'.format(value, operationStateContainer.AllowedRange))

    def _getAcSourceRange(self, source):
        """
        Given the source of the AlertCondition return PhysiologicalRange
        Only returns range with Lower, Upper and StepWidth
        @param source: Source metric handle
        @return: (Range) PhysiologicalRange
        """

        for handle in source:
            state = self._mdib.states.descriptorHandle.getOne(handle)
            ranges = state.PhysiologicalRange
            # find range that has Upper Lower and Step
            for range in ranges:
                if range.Upper is not None and range.Lower is not None and range.StepWidth is not None:
                    return range
        return None

    def _generateValueWithinRange(self, sourceRange, previousValue, isLower=True):
        """ Generates a value within range different from the previous value"""
        if sourceRange is None:
            if isLower:
                return -random.random()
            else:
                return random.random()

        exp = str(sourceRange.StepWidth)[::-1].find(".")
        if exp == -1:
            exp = 0

        value = previousValue
        for cnt in range(int((sourceRange.Upper - sourceRange.Lower) // sourceRange.StepWidth)):
            if value != previousValue: break

            if isLower:
                value = round(sourceRange.Lower + (sourceRange.StepWidth*cnt), exp)
            else:
                value = round(sourceRange.Upper - (sourceRange.StepWidth*cnt), exp)

        return value

    def _activateAllDelegableAlertSignals(self, operationDescriptorContainer, value): #pylint: disable=unused-argument
        # find all delegable Alert Signals in main alert system
        # set ActivationState to "On"
        # set all corresponding local alert signals to paused
        allAlertConditionDescriptors = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('AlertConditionDescriptor'), [])
        allAlertConditionDescriptors += self._mdib.descriptions.NODETYPE.get(namespaces.domTag('LimitAlertConditionDescriptor'), [])
        self._lastActivateAllDelegableAlerts = time.time()
        with self._mdib.mdibUpdateTransaction() as mgr:
            for acd in allAlertConditionDescriptors:
                signalDescriptors = self._mdib.descriptions.ConditionSignaled.get(acd.handle, [])
                for sd in signalDescriptors:
                    if sd.SignalDelegationSupported:
                        ss = mgr.getAlertState(sd.handle)
                        if ss.ActivationState == AlertActivation.OFF:
                            ss.ActivationState = AlertActivation.ON
                            self._pauseFallbackAlertSignals(sd, signalDescriptors, mgr)
                        else:
                            mgr.ungetState(ss)

    def _pauseFallbackAlertSignals(self, delegableSignalDescriptor, allSignalDescriptors, transaction):
        if allSignalDescriptors is None:
            allSignalDescriptors = self._mdib.descriptions.ConditionSignaled.get(delegableSignalDescriptor.ConditionSignaled, [])

        # look for local fallback signal (same Manifestation), and set it to paused
        fallbacks = [tmp for tmp in allSignalDescriptors if
                     not tmp.SignalDelegationSupported and tmp.Manifestation == delegableSignalDescriptor.Manifestation]
        for f in fallbacks:
            ss_fallback = transaction.getAlertState(f.handle)
            if ss_fallback.ActivationState != AlertActivation.PAUSED:
                ss_fallback.ActivationState = AlertActivation.PAUSED
            else:
                transaction.ungetState(ss_fallback)

    def _activateFallbackAlertSignals(self, delegableSignalDescriptor, allSignalDescriptors, transaction):
        if allSignalDescriptors is None:
            allSignalDescriptors = self._mdib.descriptions.ConditionSignaled.get(delegableSignalDescriptor.ConditionSignaled, [])

        # look for local fallback signal (same Manifestation), and set it to paused
        fallbacks = [tmp for tmp in allSignalDescriptors if
                     not tmp.SignalDelegationSupported and tmp.Manifestation == delegableSignalDescriptor.Manifestation]
        for f in fallbacks:
            ss_fallback = transaction.getAlertState(f.handle)
            if ss_fallback.ActivationState == AlertActivation.PAUSED:
                ss_fallback.ActivationState = AlertActivation.ON
            else:
                transaction.ungetState(ss_fallback)

    def _activateAlertSystem(self, operationDescriptorContainer, _):
        self._setAlertSystemActivationState(operationDescriptorContainer.operationTarget, AlertActivation.ON)

    def _deactivateAlertSystem(self, operationDescriptorContainer, _):
        self._setAlertSystemActivationState(operationDescriptorContainer.operationTarget, AlertActivation.OFF)

    def _setAlertSystemActivationState(self, handle, alertActivationState):
        """
        R0116: The activation state of pm:AlertSystemState SHALL result in an activation state of
            pm:AlertConditionState and pm:AlertSignalState
        """
        alertConditions, alertSignals = [], []
        for child in self._mdib.descriptions.parentHandle.get(handle, []):
            if child.isAlertSignalDescriptor:
                alertSignals.append(child.handle)
            elif child.isAlertConditionDescriptor:
                alertConditions.append(child.handle)

        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getAlertState(handle)
            state.ActivationState = alertActivationState
            for alertConditionHandle in alertConditions:
                self._setAlertConditionActivationState(mgr, alertConditionHandle, alertActivationState)
            for alertSignalHandle in alertSignals:
                self._setAlertSignalActivationState(mgr, alertSignalHandle, alertActivationState)

    def _activateAlertCondition(self, operationDescriptorContainer, _):
        with self._mdib.mdibUpdateTransaction() as mgr:
            self._setAlertConditionActivationState(mgr, operationDescriptorContainer.operationTarget, AlertActivation.ON)

    def _deactivateAlertCondition(self, operationDescriptorContainer, _):
        with self._mdib.mdibUpdateTransaction() as mgr:
            self._setAlertConditionActivationState(mgr, operationDescriptorContainer.operationTarget, AlertActivation.OFF)

    def _setAlertConditionActivationState(self, mgr, handle, alertActivationState):
        state = mgr.getAlertState(handle)
        state.ActivationState = alertActivationState

    def _activateAlertSignal(self, operationDescriptorContainer, _):
        operationTargetHandle = operationDescriptorContainer.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            self._setAlertSignalActivationState(mgr, operationTargetHandle, AlertActivation.ON)

    def _deactivateAlertSignal(self, operationDescriptorContainer, _):
        operationTargetHandle = operationDescriptorContainer.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            self._setAlertSignalActivationState(mgr, operationTargetHandle, AlertActivation.OFF)

    def _setAlertSignalActivationState(self, mgr, handle, alertActivationState):
        self._lastSetAlertSignalState[handle] = time.time()
        state = mgr.getAlertState(handle)
        state.ActivationState = alertActivationState
        descr = self._mdib.descriptions.handle.getOne(handle)
        if descr.SignalDelegationSupported:
            if alertActivationState == AlertActivation.ON:
                self._pauseFallbackAlertSignals(descr, None, mgr)
            else:
                self._activateFallbackAlertSignals(descr, None, mgr)

    def _setAlertSignalState(self, operationDescriptorContainer, value):
        operationTargetHandle = operationDescriptorContainer.operationTarget
        self._lastSetAlertSignalState[operationTargetHandle] = time.time()
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('set alert state {} of {} from {} to {}', operationTargetHandle, state, state.ActivationState, value.ActivationState)
            state.ActivationState = value.ActivationState
            state.Presence = value.Presence
            state.ActualSignalGenerationDelay = value.ActualSignalGenerationDelay
            descr = self._mdib.descriptions.handle.getOne(operationTargetHandle)
            if descr.SignalDelegationSupported:
                if value.ActivationState == AlertActivation.ON:
                    self._pauseFallbackAlertSignals(descr, None, mgr)
                else:
                    self._activateFallbackAlertSignals(descr, None, mgr)

    def _workerThreadLoop(self):
        # delay start of operation
        shall_stop = self._stopWorker.wait(timeout=self.WORKERTHREAD_INTERVAL)
        if shall_stop:
            return

        while True:
            shall_stop = self._stopWorker.wait(timeout=self.WORKERTHREAD_INTERVAL)
            if shall_stop:
                return
            self._updateAlertSystemState_CurrentAlerts()
            self._handleDelegateTimeouts()

    def _getAlertSystemStates_needingUpdate(self):
        '''

        :return: all AlertSystemStateContainers of those last
        '''
        alertStatesNeedingUpdate = []
        try:
            all_alert_systems_descr = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('AlertSystemDescriptor'), list())
            for alert_system_descr in all_alert_systems_descr:
                alert_system_state = self._mdib.states.descriptorHandle.getOne(alert_system_descr.handle, allowNone=True)
                selfcheck_period = alert_system_descr.SelfCheckPeriod
                if selfcheck_period is not None:
                    last_selfcheck = alert_system_state.LastSelfCheck or 0.0
                    if time.time() - last_selfcheck >= selfcheck_period:
                        alertStatesNeedingUpdate.append(alert_system_state)
        except:
            exc = traceback.format_exc()
            self._logger.error('_getAlertSystemStates_needingUpdate: {}', exc)
        return alertStatesNeedingUpdate

    def _updateAlertSystemState_CurrentAlerts(self):
        ''' updates AlertSystemState present alarms list'''
        alertStatesNeedingUpdate = self._getAlertSystemStates_needingUpdate()
        if len(alertStatesNeedingUpdate) > 0:
            try:
                with self._mdib.mdibUpdateTransaction() as tr:
                    tr_states = [tr.getAlertState(s.descriptorHandle) for s in alertStatesNeedingUpdate]
                    self._updateAlertSystemStates(self._mdib, tr, tr_states)
            except:
                exc = traceback.format_exc()
                self._logger.error('_checkAlertStates: {}', exc)

    def _handleDelegateTimeouts(self):
        if self._lastActivateAllDelegableAlerts:
            # find the minimal invocation_effective_timeout
            all_op_descrs = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('SetAlertStateOperationDescriptor'), [])
            timeouts = [op.InvocationEffectiveTimeout for op in all_op_descrs]
            timeouts = [t for t in timeouts if t is not None]
            if not timeouts:
                return # nothing to do
            minimal_invocation_effective_timeout = min(timeouts)
            if time.time() - self._lastActivateAllDelegableAlerts > minimal_invocation_effective_timeout:
                # expired, set all AlertSignalState.ActivationState to 'Off'
                with self._mdib.mdibUpdateTransaction() as mgr:
                    for op in all_op_descrs:
                        signalDescr = self._mdib.descriptions.handle.getOne(op.OperationTarget)
                        allSignalDescriptors = self._mdib.descriptions.ConditionSignaled.get(signalDescr.ConditionSignaled, [])
                        ss = mgr.getAlertState(signalDescr.handle)
                        if ss.ActivationState == AlertActivation.ON:
                            ss.ActivationState = AlertActivation.OFF
                            self._activateFallbackAlertSignals(signalDescr, allSignalDescriptors, mgr)
                        else:
                            mgr.ungetState(ss)
                self._lastActivateAllDelegableAlerts = 0



import platform

# OS dependent import
if platform.system() == 'Windows':
    from .ntifmanagerdll import getNetworkAdapterConfigs, GetAdaptersAddresses
else:
    from .posixifmanager import getNetworkAdapterConfigs, GetAdaptersAddresses



from socket import AF_INET
import ctypes
from ctypes.wintypes import DWORD, BYTE
import traceback
import struct
import socket
from collections import namedtuple

_iphlpapi = ctypes.windll.iphlpapi

#################################### structures for getAdaptersAddresses #################################

MAX_ADAPTER_ADDRESS_LENGTH = 8
MAX_DHCPV6_DUID_LENGTH = 130

class SOCKADDR(ctypes.Structure):
    _fields_ = [
        ('family', ctypes.c_ushort),
        ('data', ctypes.c_byte*14),
        ]
LPSOCKADDR = ctypes.POINTER(SOCKADDR)
class SOCKET_ADDRESS(ctypes.Structure):
    _fields_ = [
        ('address', LPSOCKADDR),
        ('length', ctypes.c_int),
        ]

class IP_ADAPTER_PREFIX(ctypes.Structure):
    pass
PIP_ADAPTER_PREFIX = ctypes.POINTER(IP_ADAPTER_PREFIX)
IP_ADAPTER_PREFIX._fields_ = [  #pylint: disable=protected-access
    ("alignment", ctypes.c_ulonglong),
    ("next", PIP_ADAPTER_PREFIX),
    ("address", SOCKET_ADDRESS),
    ("prefix_length", ctypes.c_ulong)
    ]

class IP_ADAPTER_UNICAST_ADDRESS(ctypes.Structure):
    pass
PIP_ADAPTER_UNICAST_ADDRESS = ctypes.POINTER(IP_ADAPTER_UNICAST_ADDRESS)
IP_ADAPTER_UNICAST_ADDRESS._fields_ = [ #pylint: disable=protected-access
        ("length", ctypes.c_ulong),
        ("flags", ctypes.wintypes.DWORD),
        ("next", PIP_ADAPTER_UNICAST_ADDRESS),
        ("address", SOCKET_ADDRESS),
        ("prefix_origin", ctypes.c_int),
        ("suffix_origin", ctypes.c_int),
        ("dad_state", ctypes.c_int),
        ("valid_lifetime", ctypes.c_ulong),
        ("preferred_lifetime", ctypes.c_ulong),
        ("lease_lifetime", ctypes.c_ulong),
        ("on_link_prefix_length", ctypes.c_ubyte)
        ]

class IP_ADAPTER_ADDRESSES(ctypes.Structure):
    pass
LP_IP_ADAPTER_ADDRESSES = ctypes.POINTER(IP_ADAPTER_ADDRESSES)
    
# for now, just use void * for pointers to unused structures
PIP_ADAPTER_ANYCAST_ADDRESS = ctypes.c_void_p
PIP_ADAPTER_MULTICAST_ADDRESS = ctypes.c_void_p
PIP_ADAPTER_DNS_SERVER_ADDRESS = ctypes.c_void_p
#PIP_ADAPTER_PREFIX = ctypes.c_void_p
PIP_ADAPTER_WINS_SERVER_ADDRESS_LH = ctypes.c_void_p
PIP_ADAPTER_GATEWAY_ADDRESS_LH = ctypes.c_void_p
PIP_ADAPTER_DNS_SUFFIX = ctypes.c_void_p

IF_OPER_STATUS = ctypes.c_uint # this is an enum, consider http://code.activestate.com/recipes/576415/
IF_LUID = ctypes.c_uint64

NET_IF_COMPARTMENT_ID = ctypes.c_uint32
GUID = ctypes.c_byte*16
NET_IF_NETWORK_GUID = GUID
NET_IF_CONNECTION_TYPE = ctypes.c_uint # enum
TUNNEL_TYPE = ctypes.c_uint # enum

IP_ADAPTER_ADDRESSES._fields_ = [ #pylint: disable=protected-access
    #('u', _IP_ADAPTER_ADDRESSES_U1),
        ('length', ctypes.c_ulong),
        ('interface_index', DWORD),
    ('next', LP_IP_ADAPTER_ADDRESSES),
    ('adapter_name', ctypes.c_char_p),
    ('first_unicast_address', PIP_ADAPTER_UNICAST_ADDRESS),
    ('first_anycast_address', PIP_ADAPTER_ANYCAST_ADDRESS),
    ('first_multicast_address', PIP_ADAPTER_MULTICAST_ADDRESS),
    ('first_dns_server_address', PIP_ADAPTER_DNS_SERVER_ADDRESS),
    ('dns_suffix', ctypes.c_wchar_p),
    ('description', ctypes.c_wchar_p),
    ('friendly_name', ctypes.c_wchar_p),
    ('byte', BYTE*MAX_ADAPTER_ADDRESS_LENGTH),
    ('physical_address_length', DWORD),
    ('flags', DWORD),
    ('mtu', DWORD),
    ('interface_type', DWORD),
    ('oper_status', IF_OPER_STATUS),
    ('ipv6_interface_index', DWORD),
    ('zone_indices', DWORD),
    ('first_prefix', PIP_ADAPTER_PREFIX),
    ('transmit_link_speed', ctypes.c_uint64),
    ('receive_link_speed', ctypes.c_uint64),
    ('first_wins_server_address', PIP_ADAPTER_WINS_SERVER_ADDRESS_LH),
    ('first_gateway_address', PIP_ADAPTER_GATEWAY_ADDRESS_LH),
    ('ipv4_metric', ctypes.c_ulong),
    ('ipv6_metric', ctypes.c_ulong),
    ('luid', IF_LUID),
    ('dhcpv4_server', SOCKET_ADDRESS),
    ('compartment_id', NET_IF_COMPARTMENT_ID),
    ('network_guid', NET_IF_NETWORK_GUID),
    ('connection_type', NET_IF_CONNECTION_TYPE),
    ('tunnel_type', TUNNEL_TYPE),
    ('dhcpv6_server', SOCKET_ADDRESS),
    ('dhcpv6_client_duid', ctypes.c_byte*MAX_DHCPV6_DUID_LENGTH),
    ('dhcpv6_client_duid_length', ctypes.c_ulong),
    ('dhcpv6_iaid', ctypes.c_ulong),
    ('first_dns_suffix', PIP_ADAPTER_DNS_SUFFIX),
    ]

def GetAdaptersAddresses():
    """
    Returns an iteratable list of adapters
    """ 
    size = ctypes.c_ulong()
    _GetAdaptersAddresses = _iphlpapi.GetAdaptersAddresses
    _GetAdaptersAddresses.argtypes = [
        ctypes.c_ulong,
        ctypes.c_ulong,
        ctypes.c_void_p,
        ctypes.POINTER(IP_ADAPTER_ADDRESSES),
        ctypes.POINTER(ctypes.c_ulong),
    ]
    _GetAdaptersAddresses.restype = ctypes.c_ulong
    res = _GetAdaptersAddresses(AF_INET,0,None, None,size)
    if res != 0x6f: # BUFFER OVERFLOW
        raise RuntimeError("Error getting structure length (%d)" % res)
    pointer_type = ctypes.POINTER(IP_ADAPTER_ADDRESSES)
    size.value = 50000 # reserve a lot of memory, computers with docker containers can have maaaany adapters
    tmp_buffer = ctypes.create_string_buffer(size.value)
    struct_p = ctypes.cast(tmp_buffer, pointer_type)
    res = _GetAdaptersAddresses(AF_INET,0,None, struct_p, size)
    if res != 0x0: # NO_ERROR:
        raise RuntimeError("Error retrieving table (%d)" % res)
    while struct_p:
        yield struct_p.contents
        struct_p = struct_p.contents.next


NetworkAdapterConfig = namedtuple('NetworkAdapterConfig', 'ip name friendly_name') 
# name: e.g "3Com EtherLink XL 10/100 PCI TX NIC (3C905B-TX)"
# friendly_name: e.g. "Local Area Connection 2"

def getNetworkAdapterConfigs(print_error=False):
    networkAdapters = []
    adapters = GetAdaptersAddresses()  # list all enabled adapters
    for a in adapters:
        netConnectionId = a.friendly_name
        # first check if adapter has an address. ignore it if not
        try:
            if a.first_unicast_address and a.first_unicast_address.contents:
                try:
                    fu = a.first_unicast_address.contents
                    ad = fu.address.address.contents
                    ip_int = struct.unpack('>2xI8x', ad.data)[0]
                    ip = socket.inet_ntoa(struct.pack("!I", ip_int))
                    cnf = NetworkAdapterConfig(ip,
                                               a.description,
                                               a.friendly_name)
                    networkAdapters.append(cnf)
                except:
                    if print_error:
                        print ('could not determine IP address of adapter "{}": {}'.format(netConnectionId, traceback.format_exc()))
        except:
            print('could not determine data of adapter "{}": {}'.format(netConnectionId, traceback.format_exc()))
    return networkAdapters



import netifaces


class Adapter():
    friendly_name = ""
    ip = ""


def getNetworkAdapterConfigs():
    interfaces = netifaces.interfaces()
    adapters = []
    for ad in interfaces:
        addresses = netifaces.ifaddresses(ad)
        if netifaces.AF_INET in addresses:
            adptr = Adapter()
            adptr.ip = addresses[netifaces.AF_INET][0]['addr']
            adptr.friendly_name = ad
            adapters.append(adptr)
    return adapters


def GetAdaptersAddresses():
    interfaces = netifaces.interfaces()
    adapters = []
    for ad in interfaces:
        addresses = netifaces.ifaddresses(ad)
        if netifaces.AF_INET in addresses:
            adptr = Adapter()
            adptr.ip = addresses[netifaces.AF_INET][0]['addr']
            adptr.friendly_name = ad
            adapters.append(adptr)
    return adapters



from lxml import etree as etree_
from .definitions_base import ProtocolsRegistry


# class BicepsSchema(object):
#     def __init__(self, definition_cls):
#         '''
#
#         :param base_definition_cls: a class derived from BaseDefinitions, it contains paths to xml schema files
#         '''
#         self.parser = etree_.ETCompatXMLParser()
#         self._versionRef = definition_cls
#         self.parser.resolvers.add(self._versionRef.schemaResolver)
#
#         pmSchema = self._parseFile(self._versionRef.ParticipantModelSchemaFile)
#         self.pmSchema = etree_.XMLSchema(etree=pmSchema)
#
#         bmmSchema = self._parseFile(self._versionRef.MessageModelSchemaFile)
#         self.bmmSchema = etree_.XMLSchema(etree=bmmSchema)
#
#         mexSchema = self._parseFile(self._versionRef.MetaDataExchangeSchemaFile, normalized=False)
#         self.mexSchema = etree_.XMLSchema(etree=mexSchema)
#
#         evtSchema = self._parseFile(self._versionRef.EventingSchemaFile, normalized=False)
#         self.evtSchema = etree_.XMLSchema(etree=evtSchema)
#
#         s12Schema = self._parseFile(self._versionRef.SoapEnvelopeSchemaFile, normalized=False)
#         self.s12Schema = etree_.XMLSchema(etree=s12Schema)
#
#         dpwsSchema = self._parseFile(self._versionRef.DPWSSchemaFile, normalized=False)
#         self.dpwsSchema = etree_.XMLSchema(etree=dpwsSchema)
#
#     def __str__(self):
#         return '{} {}'.format(self.__class__.__name__, self._versionRef.__name__)
#
#     def _parseFile(self, path, normalized=True):
#         with open(path, 'rb') as f:
#             xml_text = f.read()
#         if normalized:
#             xml_text = self._versionRef.normalizeXMLText(xml_text)
#         return etree_.fromstring(xml_text, parser=self.parser, base_url=path)


def _shortActionString(action):
    for cls in ProtocolsRegistry.protocols:
        if cls.ActionsNamespace is not None and action.startswith(cls.ActionsNamespace):
            return '{}:{}'.format(cls.__name__, action[len(cls.ActionsNamespace):])
    return action


def shortFilterString(actions):
    '''
    Helper function to make shorter action strings for logging
    :param actions: list of strings
    :return: a comma separated string of shortened names
    '''
    return ', '.join([_shortActionString(a) for a in actions])



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os
import subprocess

version = '1.1.15'

# create a version.py file that is
# a) used for __version__ info
# b) contains current git hash of repo
try:
    gitrev = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip()
    gitrev = gitrev.decode('ascii')
except:
    import traceback
    print (traceback.format_exc())
    gitrev = '-'

here = os.path.abspath(os.path.dirname(__file__))

# write version file
with open(os.path.join(here, 'sdc11073/version.py'), 'w') as v:
    v.write("#generated file!\nversion='{}'\n\ngitrev='''{}'''".format(version, gitrev))
    
# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

dependencies = ['lxml>=2.3',
                "netifaces ; platform_system!='Windows'"]

setup(
    name='sdc11073',
    version=version,
    description='pure python implementation of IEEE11073 SDC protocol',
    long_description=long_description,
    url='https://github.com/Draegerwerk/sdc11073',
    author='Bernd Deichmann',
    author_email='bernd.deichmann@draeger.com',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    # What does your project relate to?
    keywords='SDC IEEE11073',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(include=['sdc11073', 'sdc11073.*']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=dependencies,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
#    extras_require={
#        'dev': ['check-manifest'],
#        'test': ['coverage'],
#    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'sdc11073': ['tutorial/readme.rst',
                     'tutorial/consumer/*.py',
                     'tutorial/provider/*.xml',
                     'tutorial/provider/*.py',
                     'xsd/*.xsd',
                     'ca/*.*',
                     'codings/*.csv'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
#    data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
#    entry_points={
#        'console_scripts': [
#            'sample=sample:main',
#        ],
#    },
)



import ssl
import os

def mk_ssl_context_from_folder(ca_folder,
                               private_key='userkey.pem',
                               certificate='usercert.pem',
                               ca_public_key='cacert.pem',
                               cyphers_file=None,
                               ssl_passwd=None):
    """Convenience method for easy creation of SSL context, assuming all needed files are in the same folder.
    Create an ssl context from files 'userkey.pem', 'usercert.pem', and optional 'cacert.pem' and cyphers file
    :param ca_folder: base path of all files
    :param private_key: name of the private key file of the user
    :param certificate: name of the signed certificate of the user
    :param ca_public_key: name of public key of the certificate authority that signed the certificate; if given,
                   verify_mode of sslContext will be set to CERT_REQUIRED
    :param cyphers_file: optional file that contains a cyphers string; comments are possible, start line with '#'
    :param ssl_passwd: optional password string
    :return: SSLContext instance
    """
    certfile = os.path.join(ca_folder, certificate)
    keyfile = os.path.join(ca_folder, private_key)
    if ca_public_key:
        cafile = os.path.join(ca_folder, ca_public_key)
    else:
        cafile = None
    if cyphers_file:
        with open(os.path.join(ca_folder, cyphers_file)) as f:
            while True:
                # allow comment lines, starting with #
                cyphers = f.readline()
                if len(cyphers) == 0: # end of file reached without having found a valid line
                    cyphers = None
                    break
                cyphers = cyphers.strip()
                cyphers = cyphers.rstrip('\n')
                cyphers = cyphers.rstrip('\r')
                if len(cyphers) > 0 and not cyphers.startswith("#"):
                    break
    else:
        cyphers = None
    return mk_ssl_context(keyfile, certfile, cafile, cyphers, ssl_passwd)


def mk_ssl_context(key_file,
                   cert_file,
                   ca_file,
                   cyphers=None,
                   ssl_passwd=None):
    """Convenience method for easy creation of SSL context.
    Create an ssl context from files 'userkey.pem', 'usercert.pem', 'cacert.pem' and optional 'cyphers.json'
    :param key_file: the private key pem file of the user
    :param cert_file: the signed certificate of the user
    :param ca_file: optional public key of the certificate authority that signed the certificate; if given,
                   verify_mode of sslContext will be set to CERT_REQUIRED
    :param cyphers: optional cyphers string
    :param ssl_passwd: optional password string
    :return: SSLContext instance
    """
    ssl_context = ssl.SSLContext()  # defaults to ssl.PROTOCOL_TLS
    ssl_context.load_cert_chain(certfile=cert_file, keyfile=key_file, password=ssl_passwd)
    if cyphers is not None:
        ssl_context.set_ciphers(cyphers)
    if ca_file:
        ssl_context.verify_mode = ssl.CERT_REQUIRED
        ssl_context.load_verify_locations(ca_file)
    return ssl_context
